#!/usr/bin/env python
'''waypoint command handling'''

import time, os, fnmatch, copy, platform, struct
from pymavlink import mavutil, mavwp
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import *

try:
    # py2
    from StringIO import StringIO as SIO
except ImportError:
    # py3
    from io import BytesIO as SIO
    
class WPModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(WPModule, self).__init__(mpstate, "wp", "waypoint handling", public = True)
        self.wp_op = None
        self.wp_requested = {}
        self.wp_received = {}
        self.wp_save_filename = None
        self.wploader_by_sysid = {}
        self.loading_waypoints = False
        self.loading_waypoint_lasttime = time.time()
        self.last_waypoint = 0
        self.wp_period = mavutil.periodic_event(0.5)
        self.undo_wp = None
        self.undo_type = None
        self.undo_wp_idx = -1
        self.upload_start = None
        self.wploader.expected_count = 0
        self.last_get_home = time.time()
        self.add_command('wp', self.cmd_wp,       'waypoint management',
                         ["<list|clear|move|remove|loop|set|undo|movemulti|moverelhome|changealt|param|status|slope|ftp|add_takeoff|add_landing|add_dls|add_rtl>",
                          "<load|update|save|savecsv|show|ftpload> (FILENAME)"])
        self.mission_ftp_name = "@MISSION/mission.dat"
        self.ftp_count = None

        if self.continue_mode and self.logdir is not None:
            waytxt = os.path.join(mpstate.status.logdir, 'way.txt')
            if os.path.exists(waytxt):
                self.wploader.load(waytxt)
                print("Loaded waypoints from %s" % waytxt)

        self.menu_added_console = False
        self.menu_added_map = False
        if mp_util.has_wxpython:
            self.menu = MPMenuSubMenu('Mission',
                                  items=[MPMenuItem('Editor', 'Editor', '# wp editor'),
                                         MPMenuItem('Clear', 'Clear', '# wp clear'),
                                         MPMenuItem('List', 'List', '# wp list'),
                                         MPMenuItem('FTP', 'FTP', '# wp ftp'),
                                         MPMenuItem('Load', 'Load', '# wp load ',
                                                    handler=MPMenuCallFileDialog(flags=('open',),
                                                                                 title='Mission Load',
                                                                                 wildcard='MissionFiles(*.txt.*.wp,*.waypoints)|*.txt;*.wp;*.waypoints')),
                                         MPMenuItem('Save', 'Save', '# wp save ',
                                                    handler=MPMenuCallFileDialog(flags=('save', 'overwrite_prompt'),
                                                                                 title='Mission Save',
                                                                                 wildcard='MissionFiles(*.txt.*.wp,*.waypoints)|*.txt;*.wp;*.waypoints')),
                                         MPMenuItem('Draw', 'Draw', '# wp draw ',
                                                    handler=MPMenuCallTextDialog(title='Mission Altitude (m)',
                                                                                 default=100)),
                                         MPMenuItem('Undo', 'Undo', '# wp undo'),
                                         MPMenuItem('Loop', 'Loop', '# wp loop'),
                                         MPMenuItem('Add NoFly', 'Loop', '# wp noflyadd'),
                                         MPMenuItem('Add Takeoff', 'Add Takeoff', '# wp add_takeoff ',
                                             handler=MPMenuCallTextDialog(title='Takeoff Altitude (m)',
                                                                                 default=20)),
                                         MPMenuItem('Add Landing', 'Add Landing', '# wp add_landing'),
                                         MPMenuItem('Add RTL', 'Add RTL', '# wp add_rtl'),
                                         MPMenuItem('Add DO_LAND_START', 'Add DO_LAND_START', '# wp add_dls')])

    @property
    def wploader(self):
        '''per-sysid wploader'''
        if self.target_system not in self.wploader_by_sysid:
            self.wploader_by_sysid[self.target_system] = mavwp.MAVWPLoader()
        return self.wploader_by_sysid[self.target_system]

    def missing_wps_to_request(self):
        ret = []
        tnow = time.time()
        next_seq = self.wploader.count()
        for i in range(5):
            seq = next_seq+i
            if seq+1 > self.wploader.expected_count:
                continue
            if seq in self.wp_requested and tnow - self.wp_requested[seq] < 2:
                continue
            ret.append(seq)
        return ret

    def is_quadplane(self):
        Q_ENABLE = int(self.get_mav_param("Q_ENABLE",0))
        return Q_ENABLE > 0

    def send_wp_requests(self, wps=None):
        '''send some more WP requests'''
        if wps is None:
            wps = self.missing_wps_to_request()
        tnow = time.time()
        for seq in wps:
            self.wp_requested[seq] = tnow
            if self.settings.wp_use_mission_int:
                self.master.mav.mission_request_int_send(self.master.target_system, self.master.target_component, seq)
            else:
                self.master.mav.mission_request_send(self.master.target_system, self.master.target_component, seq)

    def wp_status(self):
        '''show status of wp download'''
        try:
            print("Have %u of %u waypoints" % (self.wploader.count()+len(self.wp_received), self.wploader.expected_count))
        except Exception:
            print("Have %u waypoints" % (self.wploader.count()+len(self.wp_received)))


    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        mtype = m.get_type()
        if mtype in ['WAYPOINT_COUNT','MISSION_COUNT']:
            self.wploader.expected_count = m.count
            if self.wp_op is None:
                #self.console.error("No waypoint load started")
                pass
            else:
                self.wploader.clear()
                self.console.writeln("Requesting %u waypoints t=%s now=%s" % (m.count,
                                                                                 time.asctime(time.localtime(m._timestamp)),
                                                                                 time.asctime()))
                self.send_wp_requests()

        elif mtype in ['WAYPOINT', 'MISSION_ITEM', 'MISSION_ITEM_INT'] and self.wp_op is not None:
            if m.get_type() == 'MISSION_ITEM_INT':
                if getattr(m, 'mission_type', 0) != 0:
                    # this is not a mission item, likely fence
                    return
                # our internal structure assumes MISSION_ITEM'''
                m = self.wp_from_mission_item_int(m)
            if m.seq < self.wploader.count():
                #print("DUPLICATE %u" % m.seq)
                return
            if m.seq+1 > self.wploader.expected_count:
                self.console.writeln("Unexpected waypoint number %u - expected %u" % (m.seq, self.wploader.count()))
            self.wp_received[m.seq] = m
            next_seq = self.wploader.count()
            while next_seq in self.wp_received:
                m = self.wp_received.pop(next_seq)
                self.wploader.add(m)
                next_seq += 1
            if self.wploader.count() != self.wploader.expected_count:
                #print("m.seq=%u expected_count=%u" % (m.seq, self.wploader.expected_count))
                self.send_wp_requests()
                return
            if self.wp_op == 'list':
                self.show_and_save(m.get_srcSystem())
                self.loading_waypoints = False
            elif self.wp_op == "save":
                self.save_waypoints(self.wp_save_filename)
            self.wp_op = None
            self.wp_requested = {}
            self.wp_received = {}

        elif mtype in ["WAYPOINT_REQUEST", "MISSION_REQUEST"]:
            self.process_waypoint_request(m, self.master)

        elif mtype in ["WAYPOINT_CURRENT", "MISSION_CURRENT"]:
            if m.seq != self.last_waypoint:
                self.last_waypoint = m.seq
                if self.settings.wpupdates:
                    self.say("waypoint %u" % m.seq,priority='message')

        elif mtype == "MISSION_ITEM_REACHED":
            wp = self.wploader.wp(m.seq)
            if wp is None:
                # should we spit out a warning?!
                # self.say("No waypoints")
                pass
            else:
                if wp.command == mavutil.mavlink.MAV_CMD_DO_LAND_START:
                    alt_offset = self.get_mav_param('ALT_OFFSET', 0)
                    if alt_offset > 0.005:
                        self.say("ALT OFFSET IS NOT ZERO passing DO_LAND_START")

    def idle_task(self):
        '''handle missing waypoints'''
        if self.wp_period.trigger():
            # cope with packet loss fetching mission
            if self.master is not None and self.master.time_since('MISSION_ITEM') >= 2 and self.wploader.count() < getattr(self.wploader,'expected_count',0):
                wps = self.missing_wps_to_request();
                print("re-requesting WPs %s" % str(wps))
                self.send_wp_requests(wps)
        if self.module('console') is not None:
            if not self.menu_added_console:
                self.menu_added_console = True
                self.module('console').add_menu(self.menu)
        else:
            self.menu_added_console = False

        if self.module('map') is not None:
            if not self.menu_added_map:
                self.menu_added_map = True
                self.module('map').add_menu(self.menu)
        else:
            self.menu_added_map = False
        if (self.master is not None and
            'HOME_POSITION' not in self.master.messages and
            time.time() - self.last_get_home > 2):
            self.master.mav.command_long_send(self.settings.target_system,
                                            0,
                                            mavutil.mavlink.MAV_CMD_GET_HOME_POSITION,
                                            0, 0, 0, 0, 0, 0, 0, 0)
            self.last_get_home = time.time()


    def wp_to_mission_item_int(self, wp):
        '''convert a MISSION_ITEM to a MISSION_ITEM_INT. We always send as MISSION_ITEM_INT
           to give cm level accuracy'''
        if wp.get_type() == 'MISSION_ITEM_INT':
            return wp
        wp_int = mavutil.mavlink.MAVLink_mission_item_int_message(wp.target_system,
                                                                  wp.target_component,
                                                                  wp.seq,
                                                                  wp.frame,
                                                                  wp.command,
                                                                  wp.current,
                                                                  wp.autocontinue,
                                                                  wp.param1,
                                                                  wp.param2,
                                                                  wp.param3,
                                                                  wp.param4,
                                                                  int(wp.x*1.0e7),
                                                                  int(wp.y*1.0e7),
                                                                  wp.z)
        return wp_int

    def wp_from_mission_item_int(self, wp):
        '''convert a MISSION_ITEM_INT to a MISSION_ITEM'''
        wp2 = mavutil.mavlink.MAVLink_mission_item_message(wp.target_system,
                                                           wp.target_component,
                                                           wp.seq,
                                                           wp.frame,
                                                           wp.command,
                                                           wp.current,
                                                           wp.autocontinue,
                                                           wp.param1,
                                                           wp.param2,
                                                           wp.param3,
                                                           wp.param4,
                                                           wp.x*1.0e-7,
                                                           wp.y*1.0e-7,
                                                           wp.z)
        # preserve srcSystem as that is used for naming waypoint file
        wp2._header.srcSystem = wp.get_srcSystem()
        wp2._header.srcComponent = wp.get_srcComponent()
        return wp2

    def process_waypoint_request(self, m, master):
        '''process a waypoint request from the master'''
        if (m.target_system != self.settings.source_system or
            m.target_component != self.settings.source_component):
            # self.console.error("Mission request is not for me")
            return
        if (not self.loading_waypoints or
            time.time() > self.loading_waypoint_lasttime + 10.0):
            self.loading_waypoints = False
            #self.console.error("not loading waypoints")
            return
        if m.seq >= self.wploader.count():
            self.console.error("Request for bad waypoint %u (max %u)" % (m.seq, self.wploader.count()))
            return
        wp = self.wploader.wp(m.seq)
        wp.target_system = self.target_system
        wp.target_component = self.target_component
        if self.settings.wp_use_mission_int:
            wp_send = self.wp_to_mission_item_int(wp)
        else:
            wp_send = wp
        self.master.mav.send(wp_send)
        self.loading_waypoint_lasttime = time.time()
        self.mpstate.console.set_status('Mission', 'Mission %u/%u' % (m.seq, self.wploader.count()-1))
        if m.seq == self.wploader.count() - 1:
            self.loading_waypoints = False
            print("Loaded %u waypoint in %.2fs" % (self.wploader.count(), time.time() - self.upload_start))
            self.console.writeln("Sent all %u waypoints" % self.wploader.count())

    def send_all_waypoints(self):
        '''send all waypoints to vehicle'''
        self.upload_start = time.time()
        self.master.waypoint_clear_all_send()
        if self.wploader.count() == 0:
            return
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.waypoint_count_send(self.wploader.count())

    def load_waypoints(self, filename):
        '''load waypoints from a file'''
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        try:
            #need to remove the leading and trailing quotes in filename
            self.wploader.load(filename.strip('"'))
        except Exception as msg:
            print("Unable to load %s - %s" % (filename, msg))
            return
        print("Loaded %u waypoints from %s" % (self.wploader.count(), filename))
        self.send_all_waypoints()

    def update_waypoints(self, filename, wpnum):
        '''update waypoints from a file'''
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        try:
            self.wploader.load(filename)
        except Exception as msg:
            print("Unable to load %s - %s" % (filename, msg))
            return
        if self.wploader.count() == 0:
            print("No waypoints found in %s" % filename)
            return
        if wpnum == -1:
            print("Loaded %u updated waypoints from %s" % (self.wploader.count(), filename))
        elif wpnum >= self.wploader.count():
            print("Invalid waypoint number %u" % wpnum)
            return
        else:
            print("Loaded updated waypoint %u from %s" % (wpnum, filename))

        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        if wpnum == -1:
            start = 0
            end = self.wploader.count()-1
        else:
            start = wpnum
            end = wpnum
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                             self.target_component,
                                                             start, end)

    def save_waypoints(self, filename):
        '''save waypoints to a file'''
        try:
            #need to remove the leading and trailing quotes in filename
            self.wploader.save(filename.strip('"'))
        except Exception as msg:
            print("Failed to save %s - %s" % (filename, msg))
            return
        print("Saved %u waypoints to %s" % (self.wploader.count(), filename))

    def save_waypoints_csv(self, filename):
        '''save waypoints to a file in a human readable CSV file'''
        try:
            #need to remove the leading and trailing quotes in filename
            self.wploader.savecsv(filename.strip('"'))
        except Exception as msg:
            print("Failed to save %s - %s" % (filename, msg))
            return
        print("Saved %u waypoints to CSV %s" % (self.wploader.count(), filename))

    def cmd_wp_move(self, args):
        '''handle wp move'''
        if len(args) != 1:
            print("usage: wp move WPNUM")
            return
        idx = int(args[0])
        if idx < 1 or idx > self.wploader.count():
            print("Invalid wp number %u" % idx)
            return
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return
        wp = self.wploader.wp(idx)

        # setup for undo
        self.undo_wp = copy.copy(wp)
        self.undo_wp_idx = idx
        self.undo_type = "move"

        (lat, lon) = latlon
        if (getattr(self.console, 'ElevationMap', None) is not None and
            wp.frame == mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT and
            self.settings.wpterrainadjust):
            alt1 = self.console.ElevationMap.GetElevation(lat, lon)
            alt2 = self.console.ElevationMap.GetElevation(wp.x, wp.y)
            if alt1 is not None and alt2 is not None:
                wp.z += alt1 - alt2
        wp.x = lat
        wp.y = lon

        wp.target_system    = self.target_system
        wp.target_component = self.target_component
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                        self.target_component,
                                                        idx, idx)
        self.wploader.set(wp, idx)
        print("Moved WP %u to %f, %f at %.1fm" % (idx, lat, lon, wp.z))


    def cmd_wp_movemulti(self, args, latlon=None):
        '''handle wp move of multiple waypoints'''
        if len(args) < 3:
            print("usage: wp movemulti WPNUM WPSTART WPEND <rotation>")
            return
        idx = int(args[0])
        if idx < 1 or idx > self.wploader.count():
            print("Invalid wp number %u" % idx)
            return
        wpstart = int(args[1])
        if wpstart < 1 or wpstart > self.wploader.count():
            print("Invalid wp number %u" % wpstart)
            return
        wpend = int(args[2])
        if wpend < 1 or wpend > self.wploader.count():
            print("Invalid wp number %u" % wpend)
            return
        if idx < wpstart or idx > wpend:
            print("WPNUM must be between WPSTART and WPEND")
            return

        # optional rotation about center point
        if len(args) > 3:
            rotation = float(args[3])
        else:
            rotation = 0

        if latlon is None:
            latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return
        wp = self.wploader.wp(idx)
        if not self.wploader.is_location_command(wp.command):
            print("WP must be a location command")
            return

        (lat, lon) = latlon
        distance = mp_util.gps_distance(wp.x, wp.y, lat, lon)
        bearing  = mp_util.gps_bearing(wp.x, wp.y, lat, lon)

        for wpnum in range(wpstart, wpend+1):
            wp = self.wploader.wp(wpnum)
            if wp is None or not self.wploader.is_location_command(wp.command):
                continue
            (newlat, newlon) = mp_util.gps_newpos(wp.x, wp.y, bearing, distance)
            if wpnum != idx and rotation != 0:
                # add in rotation
                d2 = mp_util.gps_distance(lat, lon, newlat, newlon)
                b2 = mp_util.gps_bearing(lat, lon, newlat, newlon)
                (newlat, newlon) = mp_util.gps_newpos(lat, lon, b2+rotation, d2)

            if (getattr(self.console, 'ElevationMap', None) is not None and
                wp.frame != mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT and
                self.settings.wpterrainadjust):
                alt1 = self.console.ElevationMap.GetElevation(newlat, newlon)
                alt2 = self.console.ElevationMap.GetElevation(wp.x, wp.y)
                if alt1 is not None and alt2 is not None:
                    wp.z += alt1 - alt2
            wp.x = newlat
            wp.y = newlon
            wp.target_system    = self.target_system
            wp.target_component = self.target_component
            self.wploader.set(wp, wpnum)

        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                        self.target_component,
                                                        wpstart, wpend+1)
        print("Moved WPs %u:%u to %f, %f rotation=%.1f" % (wpstart, wpend, lat, lon, rotation))


    def cmd_wp_move_rel_home(self, args, latlon=None):
        '''handle wp move to a point relative to home by dist/bearing'''
        if len(args) < 3:
            print("usage: wp moverelhome WPNUM dist bearing")
            return
        idx = int(args[0])
        if idx < 1 or idx > self.wploader.count():
            print("Invalid wp number %u" % idx)
            return
        dist = float(args[1])
        bearing = float(args[2])

        home = self.get_home()
        if home is None:
            print("Need home")
            return

        wp = self.wploader.wp(idx)
        if not self.wploader.is_location_command(wp.command):
            print("Not a nav command")
            return
        (newlat, newlon) = mp_util.gps_newpos(home.x, home.y, bearing, dist)
        wp.x = newlat
        wp.y = newlon
        wp.target_system    = self.target_system
        wp.target_component = self.target_component
        self.wploader.set(wp, idx)

        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                        self.target_component,
                                                        idx, idx+1)
        print("Moved WP %u %.1fm bearing %.1f from home" % (idx, dist, bearing))

    def cmd_wp_changealt(self, args):
        '''handle wp change target alt of multiple waypoints'''
        if len(args) < 2:
            print("usage: wp changealt WPNUM NEWALT <NUMWP>")
            return
        idx = int(args[0])
        if idx < 1 or idx > self.wploader.count():
            print("Invalid wp number %u" % idx)
            return
        newalt = float(args[1])
        if len(args) >= 3:
            count = int(args[2])
        else:
            count = 1

        for wpnum in range(idx, idx+count):
            wp = self.wploader.wp(wpnum)
            if not self.wploader.is_location_command(wp.command):
                continue
            wp.z = newalt
            wp.target_system    = self.target_system
            wp.target_component = self.target_component
            self.wploader.set(wp, wpnum)

        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                        self.target_component,
                                                        idx, idx+count)
        print("Changed alt for WPs %u:%u to %f" % (idx, idx+(count-1), newalt))

    def cmd_wp_remove(self, args):
        '''handle wp remove'''
        if len(args) != 1:
            print("usage: wp remove WPNUM")
            return
        idx = int(args[0])
        if idx < 0 or idx >= self.wploader.count():
            print("Invalid wp number %u" % idx)
            return
        wp = self.wploader.wp(idx)

        # setup for undo
        self.undo_wp = copy.copy(wp)
        self.undo_wp_idx = idx
        self.undo_type = "remove"

        self.wploader.remove(wp)
        self.fix_jumps(idx, -1)
        self.send_all_waypoints()
        print("Removed WP %u" % idx)

    def cmd_wp_undo(self):
        '''handle wp undo'''
        if self.undo_wp_idx == -1 or self.undo_wp is None:
            print("No undo information")
            return
        wp = self.undo_wp
        if self.undo_type == 'move':
            wp.target_system    = self.target_system
            wp.target_component = self.target_component
            self.loading_waypoints = True
            self.loading_waypoint_lasttime = time.time()
            self.master.mav.mission_write_partial_list_send(self.target_system,
                                                            self.target_component,
                                                            self.undo_wp_idx, self.undo_wp_idx)
            self.wploader.set(wp, self.undo_wp_idx)
            print("Undid WP move")
        elif self.undo_type == 'remove':
            self.wploader.insert(self.undo_wp_idx, wp)
            self.fix_jumps(self.undo_wp_idx, 1)
            self.send_all_waypoints()
            print("Undid WP remove")
        else:
            print("bad undo type")
        self.undo_wp = None
        self.undo_wp_idx = -1

    def cmd_wp_param(self, args):
        '''handle wp parameter change'''
        if len(args) < 2:
            print("usage: wp param WPNUM PNUM <VALUE>")
            return
        idx = int(args[0])
        if idx < 1 or idx > self.wploader.count():
            print("Invalid wp number %u" % idx)
            return
        wp = self.wploader.wp(idx)
        param = [wp.param1, wp.param2, wp.param3, wp.param4]
        pnum = int(args[1])
        if pnum < 1 or pnum > 4:
            print("Invalid param number %u" % pnum)
            return

        if len(args) == 2:
            print("Param %u: %f" % (pnum, param[pnum-1]))
            return

        param[pnum-1] = float(args[2])
        wp.param1 = param[0]
        wp.param2 = param[1]
        wp.param3 = param[2]
        wp.param4 = param[3]

        wp.target_system    = self.target_system
        wp.target_component = self.target_component
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                        self.target_component,
                                                        idx, idx)
        self.wploader.set(wp, idx)
        print("Set param %u for %u to %f" % (pnum, idx, param[pnum-1]))

    def cmd_clear(self, args):
        '''clear waypoints'''
        clear_type = mavutil.mavlink.MAV_MISSION_TYPE_MISSION
        if len(args) > 0 and args[0] == "all":
            clear_type = mavutil.mavlink.MAV_MISSION_TYPE_ALL
        self.master.mav.mission_clear_all_send(self.target_system, self.target_component, clear_type)
        self.wploader.clear()
        self.loading_waypoint_lasttime = time.time()

    def cmd_wp(self, args):
        '''waypoint commands'''
        usage = "usage: wp <editor|list|load|update|save|set|clear|loop|remove|move|movemulti|changealt|ftp|ftpload|add_takeoff|add_landing|add_dls>"
        if len(args) < 1:
            print(usage)
            return

        if args[0] == "load":
            if len(args) != 2:
                print("usage: wp load <filename>")
                return
            self.load_waypoints(args[1])
        elif args[0] == "update":
            if len(args) < 2:
                print("usage: wp update <filename> <wpnum>")
                return
            if len(args) == 3:
                wpnum = int(args[2])
            else:
                wpnum = -1
            self.update_waypoints(args[1], wpnum)
        elif args[0] == "list":
            self.wp_op = "list"
            self.master.waypoint_request_list_send()
        elif args[0] == "save":
            if len(args) != 2:
                print("usage: wp save <filename>")
                return
            self.wp_save_filename = args[1]
            self.wp_op = "save"
            self.master.waypoint_request_list_send()
        elif args[0] == "savecsv":
            if len(args) != 2:
                print("usage: wp savecsv <filename.csv>")
                return
            self.savecsv(args[1])
        elif args[0] == "savelocal":
            if len(args) != 2:
                print("usage: wp savelocal <filename>")
                return
            self.wploader.save(args[1])
        elif args[0] == "show":
            if len(args) != 2:
                print("usage: wp show <filename>")
                return
            self.wploader.load(args[1])
        elif args[0] == "move":
            self.cmd_wp_move(args[1:])
        elif args[0] == "movemulti":
            self.cmd_wp_movemulti(args[1:], None)
        elif args[0] == "moverelhome":
            self.cmd_wp_move_rel_home(args[1:], None)
        elif args[0] == "changealt":
            self.cmd_wp_changealt(args[1:])
        elif args[0] == "param":
            self.cmd_wp_param(args[1:])
        elif args[0] == "remove":
            self.cmd_wp_remove(args[1:])
        elif args[0] == "undo":
            self.cmd_wp_undo()
        elif args[0] == "set":
            if len(args) != 2:
                print("usage: wp set <wpindex>")
                return
            self.master.waypoint_set_current_send(int(args[1]))
        elif args[0] == "split":
            self.cmd_split(args[1:])
        elif args[0] == "clear":
            self.cmd_clear(args[1:])
        elif args[0] == "editor":
            if self.module('misseditor'):
                self.mpstate.functions.process_stdin("module reload misseditor", immediate=True)
            else:
                self.mpstate.functions.process_stdin("module load misseditor", immediate=True)
        elif args[0] == "draw":
            if not 'draw_lines' in self.mpstate.map_functions:
                print("No map drawing available")
                return
            if self.get_home() is None:
                print("Need home location - please run gethome")
                return
            if len(args) > 1:
                self.settings.wpalt = int(args[1])
            self.mpstate.map_functions['draw_lines'](self.wp_draw_callback)
            print("Drawing waypoints on map at altitude %d" % self.settings.wpalt)
        elif args[0] == "sethome":
            self.set_home_location()
        elif args[0] == "loop":
            self.wp_loop()
        elif args[0] == "noflyadd":
            self.nofly_add()
        elif args[0] == "status":
            self.wp_status()
        elif args[0] == "slope":
            self.wp_slope(args[1:])
        elif args[0] == "ftp":
            self.wp_ftp_download()
        elif args[0] == "ftpload":
            if len(args) != 2:
                print("usage: wp ftpload <filename>")
                return
            self.wp_ftp_upload(args[1])
        elif args[0] == "add_takeoff":
            self.wp_add_takeoff(args[1:])
        elif args[0] == "add_landing":
            self.wp_add_landing()
        elif args[0] == "add_rtl":
            self.wp_add_RTL()
        elif args[0] == "add_dls":
            self.wp_add_dls()
        else:
            print(usage)

    def pretty_enum_value(self, enum_name, enum_value):
        if enum_name == "MAV_FRAME":
            if enum_value == 0:
                return "Abs"
            elif enum_value == 1:
                return "Local"
            elif enum_value == 2:
                return "Mission"
            elif enum_value == 3:
                return "Rel"
            elif enum_value == 4:
                return "Local ENU"
            elif enum_value == 5:
                return "Global (INT)"
            elif enum_value == 10:
                return "AGL"
        ret = mavutil.mavlink.enums[enum_name][enum_value].name
        ret = ret[len(enum_name)+1:]
        return ret

    def csv_line(self, line):
        '''turn a list of values into a CSV line'''
        self.csv_sep = ","
        return self.csv_sep.join(['"' + str(x) + '"' for x in line])

    def pretty_parameter_value(self, value):
        '''pretty parameter value'''
        return value

    def savecsv(self, filename):
        '''save waypoints to a file in human-readable CSV file'''
        f = open(filename, mode='w')
        headers = ["Seq", "Frame", "Cmd", "P1", "P2", "P3", "P4", "X", "Y", "Z"]
        print(self.csv_line(headers))
        f.write(self.csv_line(headers) + "\n")
        for w in self.wploader.wpoints:
            if getattr(w, 'comment', None):
#                f.write("# %s\n" % w.comment)
                pass
            out_list = [ w.seq,
                         self.pretty_enum_value('MAV_FRAME', w.frame),
                         self.pretty_enum_value('MAV_CMD', w.command),
                         self.pretty_parameter_value(w.param1),
                         self.pretty_parameter_value(w.param2),
                         self.pretty_parameter_value(w.param3),
                         self.pretty_parameter_value(w.param4),
                         self.pretty_parameter_value(w.x),
                         self.pretty_parameter_value(w.y),
                         self.pretty_parameter_value(w.z),
                         ]
            print(self.csv_line(out_list))
            f.write(self.csv_line(out_list) + "\n")
        f.close()

    def fetch(self):
        """Download wpts from vehicle (this operation is public to support other modules)"""
        if self.wp_op is None:  # If we were already doing a list or save, just restart the fetch without changing the operation
            self.wp_op = "fetch"
        self.master.waypoint_request_list_send()

    def wp_ftp_download(self):
        '''Download wpts from vehicle with ftp'''
        ftp = self.mpstate.module('ftp')
        if ftp is None:
            print("Need ftp module")
            return
        print("Fetching mission with ftp")
        self.ftp_count = None
        ftp.cmd_get([self.mission_ftp_name], callback=self.ftp_callback, callback_progress=self.ftp_callback_progress)

    def ftp_callback_progress(self, fh, total_size):
        '''progress callback from ftp fetch of mission'''
        if self.ftp_count is None and total_size >= 10:
            ofs = fh.tell()
            fh.seek(0)
            buf = fh.read(10)
            fh.seek(ofs)
            magic2,dtype,options,start,num_items = struct.unpack("<HHHHH", buf)
            if magic2 == 0x763d:
                self.ftp_count = num_items
        if self.ftp_count is not None:
            mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
            item_size = mavmsg.unpacker.size
            done = (total_size - 10) // item_size
            self.mpstate.console.set_status('Mission', 'Mission %u/%u' % (done, self.ftp_count))

    def ftp_callback(self, fh):
        '''callback from ftp fetch of mission'''
        if fh is None:
            print("mission: failed ftp download")
            return
        magic = 0x763d
        data = fh.read()
        magic2,dtype,options,start,num_items = struct.unpack("<HHHHH", data[0:10])
        if magic != magic2:
            print("mission: bad magic 0x%x expected 0x%x" % (magic2, magic))
            return
        if dtype != mavutil.mavlink.MAV_MISSION_TYPE_MISSION:
            print("mission: bad data type %u" % dtype)
            return

        self.wploader.clear()

        data = data[10:]
        mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
        item_size = mavmsg.unpacker.size
        while len(data) >= item_size:
            mdata = data[:item_size]
            data = data[item_size:]
            msg = mavmsg.unpacker.unpack(mdata)
            tlist = list(msg)
            t = tlist[:]
            for i in range(0, len(tlist)):
                tlist[i] = t[mavmsg.orders[i]]
            t = tuple(tlist)
            w = mavmsg(*t)
            w = self.wp_from_mission_item_int(w)
            self.wploader.add(w)
        self.show_and_save(self.target_system)


    def show_and_save(self, source_system):
        '''display waypoints and save'''
        for i in range(self.wploader.count()):
            w = self.wploader.wp(i)
            print("%u %u %.10f %.10f %f p1=%.1f p2=%.1f p3=%.1f p4=%.1f cur=%u auto=%u" % (
                w.command, w.frame, w.x, w.y, w.z,
                w.param1, w.param2, w.param3, w.param4,
                w.current, w.autocontinue))
        if self.logdir is not None:
            fname = 'way.txt'
            if source_system != 1:
                fname = 'way_%u.txt' % source_system
            waytxt = os.path.join(self.logdir, fname)
            self.save_waypoints(waytxt)
            print("Saved waypoints to %s" % waytxt)

    def wp_ftp_upload(self, filename):
        '''upload waypoints to vehicle with ftp'''
        ftp = self.mpstate.module('ftp')
        if ftp is None:
            print("Need ftp module")
            return
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        try:
            #need to remove the leading and trailing quotes in filename
            self.wploader.load(filename.strip('"'))
        except Exception as msg:
            print("Unable to load %s - %s" % (filename, msg))
            return
        print("Loaded %u waypoints from %s" % (self.wploader.count(), filename))
        print("Sending mission with ftp")

        fh = SIO()
        fh.write(struct.pack("<HHHHH", 0x763d,mavutil.mavlink.MAV_MISSION_TYPE_MISSION,0,0,self.wploader.count()))
        mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
        for i in range(self.wploader.count()):
            w = self.wploader.wp(i)
            w = self.wp_to_mission_item_int(w)
            tlist = []
            for field in mavmsg.ordered_fieldnames:
                tlist.append(getattr(w, field))
            tlist = tuple(tlist)
            buf = mavmsg.unpacker.pack(*tlist)
            fh.write(buf)
        fh.seek(0)

        self.upload_start = time.time()

        ftp.cmd_put([self.mission_ftp_name, self.mission_ftp_name],
                    fh=fh, callback=self.ftp_upload_callback, progress_callback=self.ftp_upload_progress)

    def ftp_upload_progress(self, proportion):
        '''callback from ftp put of mission'''
        if proportion is None:
            self.mpstate.console.set_status('Mission', 'Mission ERR')
        else:
            count = self.wploader.count()
            self.mpstate.console.set_status('Mission', 'Mission %u/%u' % (int(proportion*count), count))

    def ftp_upload_callback(self, dlen):
        '''callback from ftp put of mission'''
        if dlen is None:
            print("Failed to send waypoints")
        else:
            mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
            item_size = mavmsg.unpacker.size
            print("Sent mission of length %u in %.2fs" % ((dlen - 10) // item_size, time.time() - self.upload_start))

    # waypoint-specific methods:
    def wp_slope(self, args):
        '''show slope of waypoints'''
        if len(args) == 2:
            # specific waypoints
            wp1 = int(args[0])
            wp2 = int(args[1])
            w1 = self.wploader.wp(wp1)
            w2 = self.wploader.wp(wp2)
            delta_alt = w1.z - w2.z
            if delta_alt == 0:
                slope = "Level"
            else:
                delta_xy = mp_util.gps_distance(w1.x, w1.y, w2.x, w2.y)
                slope = "%.1f" % (delta_xy / delta_alt)
            print("wp%u -> wp%u %s" % (wp1, wp2, slope))
            return
        if len(args) != 0:
            print("Usage: wp slope WP1 WP2")
            return
        last_w = None
        for i in range(1, self.wploader.count()):
            w = self.wploader.wp(i)
            if w.command not in [mavutil.mavlink.MAV_CMD_NAV_WAYPOINT, mavutil.mavlink.MAV_CMD_NAV_LAND]:
                continue
            if last_w is not None:
                if last_w.frame != w.frame:
                    print("WARNING: frame change %u -> %u at %u" % (last_w.frame, w.frame, i))
                delta_alt = last_w.z - w.z
                if delta_alt == 0:
                    slope = "Level"
                else:
                    delta_xy = mp_util.gps_distance(w.x, w.y, last_w.x, last_w.y)
                    slope = "%.1f" % (delta_xy / delta_alt)
                print("WP%u: slope %s" % (i, slope))
            last_w = w

    def get_default_frame(self):
        '''default frame for waypoints'''
        if self.settings.terrainalt == 'Auto':
            if self.get_mav_param('TERRAIN_FOLLOW',0) == 1:
                return mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT
            return mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT
        if self.settings.terrainalt == 'True':
            return mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT
        return mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT

    def get_home(self):
        '''get home location'''
        if 'HOME_POSITION' in self.master.messages:
            h = self.master.messages['HOME_POSITION']
            return mavutil.mavlink.MAVLink_mission_item_message(self.target_system,
                                                                self.target_component,
                                                                0,
                                                                0,
                                                                mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
                                                                0, 0, 0, 0, 0, 0,
                                                                h.latitude*1.0e-7, h.longitude*1.0e-7, h.altitude*1.0e-3)
        if self.wploader.count() > 0:
            return self.wploader.wp(0)
        return None
        

    def wp_draw_callback(self, points):
        '''callback from drawing waypoints'''
        if len(points) < 3:
            return
        from MAVProxy.modules.lib import mp_util
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        if self.wploader.count() < 2:
            home = self.get_home()
            if home is None:
                print("Need home location for draw - please run gethome")
                return
            self.wploader.clear()
            self.wploader.add(home)
        if self.get_default_frame() == mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT:
            use_terrain = True
        else:
            use_terrain = False
        for p in points:
            self.wploader.add_latlonalt(p[0], p[1], self.settings.wpalt, terrain_alt=use_terrain)
        self.send_all_waypoints()

    def wp_loop(self):
        '''close the loop on a mission'''
        loader = self.wploader
        if loader.count() < 2:
            print("Not enough waypoints (%u)" % loader.count())
            return
        wp = loader.wp(loader.count()-2)
        if wp.command == mavutil.mavlink.MAV_CMD_DO_JUMP:
            print("Mission is already looped")
            return
        if loader.count() > 1 and loader.wp(1).command in [mavutil.mavlink.MAV_CMD_NAV_TAKEOFF, mavutil.mavlink.MAV_CMD_NAV_VTOL_TAKEOFF]:
            target = 2
        else:
            target = 1
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0, 0, mavutil.mavlink.MAV_CMD_DO_JUMP,
                                                          0, 1, target, -1, 0, 0, 0, 0, 0)
        loader.add(wp)
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.waypoint_count_send(self.wploader.count())
        print("Closed loop on mission")

    def nofly_add(self):
        '''add a square flight exclusion zone'''
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position chosen")
            return
        loader = self.wploader
        (center_lat, center_lon) = latlon
        points = []
        points.append(mp_util.gps_offset(center_lat, center_lon, -25,  25))
        points.append(mp_util.gps_offset(center_lat, center_lon,  25,  25))
        points.append(mp_util.gps_offset(center_lat, center_lon,  25, -25))
        points.append(mp_util.gps_offset(center_lat, center_lon, -25, -25))
        start_idx = loader.count()
        for p in points:
            wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0, 0, mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_EXCLUSION,
                                                              0, 1, 4, 0, 0, 0, p[0], p[1], 0)
            loader.add(wp)
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                        self.target_component,
                                                        start_idx, start_idx+4)
        print("Added nofly zone")

    def wp_add_takeoff(self, args):
        '''add a takeoff as first mission item'''
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position chosen")
            return
        takeoff_alt = 20
        if len(args) > 0:
            takeoff_alt = float(args[0])
        if self.is_quadplane():
            wptype = mavutil.mavlink.MAV_CMD_NAV_VTOL_TAKEOFF
        else:
            wptype = mavutil.mavlink.MAV_CMD_NAV_TAKEOFF
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          wptype,
                                                          0, 1, 0, 0, 0, 0, latlon[0], latlon[1], takeoff_alt)
        if self.wploader.count() < 2:
            home = self.get_home()
            if home is None:
                print("Need home location - please run gethome")
                return
            self.wploader.clear()
            self.wploader.add(home)
        # assume first waypoint
        self.wploader.insert(1, wp)
        self.send_all_waypoints()

    def wp_add_landing(self):
        '''add a landing as last mission item'''
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position chosen")
            return
        if self.is_quadplane():
            wptype = mavutil.mavlink.MAV_CMD_NAV_VTOL_LAND
        else:
            wptype = mavutil.mavlink.MAV_CMD_NAV_LAND
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          wptype,
                                                          0, 1, 0, 0, 0, 0, latlon[0], latlon[1], 0)
        # assume last waypoint
        self.wploader.add(wp)
        self.send_all_waypoints()

    def wp_add_RTL(self):
        '''add a RTL as last mission item'''
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          mavutil.mavlink.MAV_CMD_NAV_RETURN_TO_LAUNCH,
                                                          0, 1, 0, 0, 0, 0, 0, 0, 0)
        # assume last waypoint
        self.wploader.add(wp)
        self.send_all_waypoints()

    def wp_add_dls(self):
        '''add a DO_LAND_START as last mission item'''
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position chosen")
            return
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          mavutil.mavlink.MAV_CMD_DO_LAND_START,
                                                          0, 1, 0, 0, 0, 0, latlon[0], latlon[1], 0)
        # assume last waypoint
        self.wploader.add(wp)
        self.send_all_waypoints()

    def set_home_location(self):
        '''set home location from last map click'''
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position available")
            return
        lat = float(latlon[0])
        lon = float(latlon[1])
        if self.wploader.count() == 0:
            self.wploader.add_latlonalt(lat, lon, 0)
        w = self.wploader.wp(0)
        w.x = lat
        w.y = lon
        self.wploader.set(w, 0)
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                             self.target_component,
                                                             0, 0)

    def fix_jumps(self, idx, delta):
        '''fix up jumps when we add/remove rows'''
        numrows = self.wploader.count()
        for row in range(numrows):
            wp = self.wploader.wp(row)
            jump_cmds = [mavutil.mavlink.MAV_CMD_DO_JUMP]
            if hasattr(mavutil.mavlink, "MAV_CMD_DO_CONDITION_JUMP"):
                jump_cmds.append(mavutil.mavlink.MAV_CMD_DO_CONDITION_JUMP)
            if wp.command in jump_cmds:
                p1 = int(wp.param1)
                if p1 > idx and p1+delta>0:
                    wp.param1 = float(p1+delta)
                    self.wploader.set(wp, row)

    def get_loc(self, m):
        '''return a mavutil.location for item m'''
        t = m.get_type()
        if t == "MISSION_ITEM":
            lat = m.x * 1e7
            lng = m.y * 1e7
            alt = m.z * 1e2
        elif t == "MISSION_ITEM_INT":
            lat = m.x
            lng = m.y
            alt = m.z
        else:
            return None
        return mavutil.location(lat, lng, alt)

    def cmd_split(self, args):
        '''splits the segment ended by the supplied waypoint into two'''
        try:
            num = int(args[0])
        except IOError as e:
            return "Bad wp num (%s)" % args[0]

        if num < 1 or num > self.wploader.count():
            print("Bad item %s" % str(num))
            return
        wp = self.wploader.wp(num)
        if wp is None:
            print("Could not get wp %u" % num)
            return
        loc = self.get_loc(wp)
        if loc is None:
            print("wp is not a location command")
            return

        prev = num - 1
        if prev < 1 or prev > self.wploader.count():
            print("Bad item %u" % num)
            return
        prev_wp = self.wploader.wp(prev)
        if prev_wp is None:
            print("Could not get previous wp %u" % prev)
            return
        prev_loc = self.get_loc(prev_wp)
        if prev_loc is None:
            print("previous wp is not a location command")
            return

        if wp.frame != prev_wp.frame:
            print("waypoints differ in frame (%u vs %u)" %
                  (wp.frame, prev_wp.frame))
            return

        if wp.frame != prev_wp.frame:
            print("waypoints differ in frame")
            return

        lat_avg = (loc.lat + prev_loc.lat)/2
        lng_avg = (loc.lng + prev_loc.lng)/2
        alt_avg = (loc.alt + prev_loc.alt)/2
        new_wp = mavutil.mavlink.MAVLink_mission_item_message(
            self.target_system,
            self.target_component,
            wp.seq,    # seq
            wp.frame,    # frame
            mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,    # command
            0,    # current
            0,    # autocontinue
            0.0,  # param1,
            0.0,  # param2,
            0.0,  # param3
            0.0,  # param4
            lat_avg * 1e-7,  # x (latitude)
            lng_avg * 1e-7,  # y (longitude)
            alt_avg * 1e-2,  # z (altitude)
        )
        self.wploader.insert(wp.seq, new_wp)
        self.fix_jumps(wp.seq, 1)
        self.send_all_waypoints()


def init(mpstate):
    '''initialise module'''
    return WPModule(mpstate)



#!/usr/bin/env python

"""
  MAVProxy sailing dashboard
"""

from MAVProxy.modules.lib import multiproc
import time

class SailingDashboard(object):
    '''
    A sailing dashboard for MAVProxy
    '''

    def __init__(self, title="MAVProxy: Sailing Dashboard"):
        self.title = title

        # create a pipe for communication from the module to the GUI
        self.child_pipe_recv, self.parent_pipe_send = multiproc.Pipe(duplex=False)
        self.close_event = multiproc.Event()
        self.close_event.clear()

        # create and start the child process
        self.child = multiproc.Process(target=self.child_task)
        self.child.start()

        # prevent the parent from using the child connection
        self.child_pipe_recv.close()

    def child_task(self):
        '''The child process hosts the GUI elements'''

        # prevent the child from using the parent connection
        self.parent_pipe_send.close()

        from MAVProxy.modules.lib import wx_processguard
        from MAVProxy.modules.lib.wx_loader import wx
        from MAVProxy.modules.lib.wxsaildash_ui import SailingDashboardFrame 

        # create the wx application and pass self as the state
        app = wx.App()
        app.frame = SailingDashboardFrame(state=self, title=self.title, size=(800, 300))
        app.frame.SetDoubleBuffered(True)
        app.frame.Show()
        app.MainLoop()

        # trigger a close event when the main app window is closed.
        # the event is monitored by the MAVProxy module which will
        # flag the module for unloading
        self.close_event.set()

    def close(self):
        '''Close the GUI'''

        # trigger a close event which is monitored by the 
        # child gui process - it will close allowing the 
        # process to be joined
        self.close_event.set()
        if self.is_alive():
            self.child.join(timeout=2.0)

    def is_alive(self):
        '''Check if the GUI process is alive'''

        return self.child.is_alive()

if __name__ == "__main__":
    '''A stand alone test for the sailing dashboard'''

    multiproc.freeze_support()
    sail_dash = SailingDashboard()
    while sail_dash.is_alive():
        print('sailing dashboard is alive')
        time.sleep(0.5)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# generated by wxGlade 0.6.8 on Wed Jun 11 13:41:49 2014
#

from ..lib.wx_loader import wx
from wx import grid
import traceback

# begin wxGlade: dependencies
# end wxGlade

import time, math, os
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import win_layout
from MAVProxy.modules.mavproxy_misseditor import me_event
MissionEditorEvent = me_event.MissionEditorEvent

from MAVProxy.modules.mavproxy_misseditor import me_defines
from MAVProxy.modules.mavproxy_map import mp_elevation

from MAVProxy.modules.mavproxy_misseditor import button_renderer

#define column names via "enums":
ME_COMMAND_COL = 0
ME_P1_COL = 1
ME_P2_COL = 2
ME_P3_COL = 3
ME_P4_COL = 4
ME_LAT_COL = 5
ME_LON_COL = 6
ME_ALT_COL = 7
ME_FRAME_COL = 8
ME_DELETE_COL = 9
ME_UP_COL = 10
ME_DOWN_COL = 11
ME_DIST_COL = 12
ME_ANGLE_COL = 13

class MissionEditorFrame(wx.Frame):
    def __init__(self, state, *args, **kwds):
        # begin wxGlade: MissionEditorFrame.__init__
        self.state = state
        kwds["style"] = wx.DEFAULT_FRAME_STYLE
        wx.Frame.__init__(self, *args, **kwds)
        self.label_sync_state = wx.StaticText(self, wx.ID_ANY, "UNSYNCED   \n", style=wx.ALIGN_CENTRE)
        self.label_wp_radius = wx.StaticText(self, wx.ID_ANY, "WP Radius")
        self.text_ctrl_wp_radius = wx.TextCtrl(self, wx.ID_ANY, "", style=wx.TE_PROCESS_ENTER | wx.TE_PROCESS_TAB)
        self.label_loiter_rad = wx.StaticText(self, wx.ID_ANY, "Loiter Radius")
        self.text_ctrl_loiter_radius = wx.TextCtrl(self, wx.ID_ANY, "", style=wx.TE_PROCESS_ENTER | wx.TE_PROCESS_TAB)
        self.checkbox_loiter_dir = wx.CheckBox(self, wx.ID_ANY, "CW")
        self.checkbox_agl = wx.CheckBox(self, wx.ID_ANY, "AGL")
        self.label_default_alt = wx.StaticText(self, wx.ID_ANY, "Default Alt")
        self.text_ctrl_wp_default_alt = wx.TextCtrl(self, wx.ID_ANY, "", style=wx.TE_PROCESS_ENTER | wx.TE_PROCESS_TAB)
        self.label_home_location = wx.StaticText(self, wx.ID_ANY, "Home Location")
        self.label_home_lat = wx.StaticText(self, wx.ID_ANY, "Lat")
        self.label_home_lat_value = wx.StaticText(self, wx.ID_ANY, "0.0")
        self.label_home_lon = wx.StaticText(self, wx.ID_ANY, "Lon")
        self.label_home_lon_value = wx.StaticText(self, wx.ID_ANY, "0.0")
        self.label_home_alt = wx.StaticText(self, wx.ID_ANY, "Alt (abs)")
        self.label_home_alt_value = wx.StaticText(self, wx.ID_ANY, "0.0")
        self.button_read_wps = wx.Button(self, wx.ID_ANY, "Read WPs")
        self.button_write_wps = wx.Button(self, wx.ID_ANY, "Write WPs")
        self.button_load_wp_file = wx.Button(self, wx.ID_ANY, "Load WP File")
        self.button_save_wp_file = wx.Button(self, wx.ID_ANY, "Save WP File")
        self.grid_mission = wx.grid.Grid(self, wx.ID_ANY, size=(1, 1))
        self.button_add_wp = wx.Button(self, wx.ID_ANY, "Add Below")
        self.button_split = wx.Button(self, wx.ID_ANY, "Split")

        self.__set_properties()
        self.__do_layout()

        self.ElevationModel = mp_elevation.ElevationModel()


        self.Bind(wx.EVT_TEXT_ENTER, self.on_wp_radius_enter, self.text_ctrl_wp_radius)
        self.Bind(wx.EVT_TEXT, self.on_wp_radius_changed, self.text_ctrl_wp_radius)
        self.Bind(wx.EVT_TEXT_ENTER, self.on_loiter_rad_enter, self.text_ctrl_loiter_radius)
        self.Bind(wx.EVT_TEXT, self.on_loiter_rad_change, self.text_ctrl_loiter_radius)
        self.Bind(wx.EVT_CHECKBOX, self.on_loiter_dir_cb_change, self.checkbox_loiter_dir)
        self.Bind(wx.EVT_TEXT_ENTER, self.on_wp_default_alt_enter, self.text_ctrl_wp_default_alt)
        self.Bind(wx.EVT_TEXT, self.on_wp_default_alt_change, self.text_ctrl_wp_default_alt)
        self.Bind(wx.EVT_BUTTON, self.read_wp_pushed, self.button_read_wps)
        self.Bind(wx.EVT_BUTTON, self.write_wp_pushed, self.button_write_wps)
        self.Bind(wx.EVT_BUTTON, self.load_wp_file_pushed, self.button_load_wp_file)
        self.Bind(wx.EVT_BUTTON, self.save_wp_file_pushed, self.button_save_wp_file)
        self.Bind(wx.grid.EVT_GRID_CMD_CELL_CHANGED, self.on_mission_grid_cell_changed, self.grid_mission)
        self.Bind(wx.grid.EVT_GRID_CMD_CELL_LEFT_CLICK, self.on_mission_grid_cell_left_click, self.grid_mission)
        self.Bind(wx.grid.EVT_GRID_CMD_SELECT_CELL, self.on_mission_grid_cell_select, self.grid_mission)
        self.Bind(wx.EVT_BUTTON, self.add_wp_below_pushed, self.button_add_wp)
        self.Bind(wx.EVT_BUTTON, self.split_pushed, self.button_split)
        # end wxGlade

        #use a timer to facilitate event an event handlers for events
        #passed from another process.
        self.timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.time_to_process_gui_events, self.timer)
        self.timer.Start(200)

        self.last_layout_send = time.time()
        self.Bind(wx.EVT_IDLE, self.on_idle)

        delete_br = button_renderer.ButtonRenderer("Delete",70,20)
        up_br = button_renderer.ButtonRenderer("+",20,20)
        down_br = button_renderer.ButtonRenderer("-",1,1)

        self.del_attr = wx.grid.GridCellAttr()
        self.del_attr.SetReadOnly(True)
        self.del_attr.SetRenderer(delete_br)

        self.up_attr = wx.grid.GridCellAttr()
        self.up_attr.SetReadOnly(True)
        self.up_attr.SetRenderer(up_br)

        self.down_attr = wx.grid.GridCellAttr()
        self.down_attr.SetReadOnly(True)
        self.down_attr.SetRenderer(down_br)
        self.read_only_attr = wx.grid.GridCellAttr()
        self.read_only_attr.SetReadOnly(True)

        self.grid_mission.SetColAttr(ME_DELETE_COL, self.del_attr)
        self.grid_mission.SetColAttr(ME_UP_COL, self.up_attr)
        self.grid_mission.SetColAttr(ME_DOWN_COL, self.down_attr)
        self.grid_mission.SetColAttr(ME_DIST_COL, self.read_only_attr)
        self.grid_mission.SetColAttr(ME_ANGLE_COL, self.read_only_attr)
        self.grid_mission.SetRowLabelSize(50)

        #remember what mission we opened/saved last
        self.last_mission_file_path = ""

        #remember last map click position
        self.last_map_click_pos = None

    def __set_properties(self):
        # begin wxGlade: MissionEditorFrame.__set_properties
        self.SetTitle("Mission Editor")
        self.SetSize((820, 480))
        self.label_sync_state.SetForegroundColour(wx.Colour(255, 0, 0))
        self.label_sync_state.SetFont(wx.Font(14, wx.DEFAULT, wx.NORMAL, wx.NORMAL, 0, "Droid Sans"))
        self.text_ctrl_wp_radius.SetMinSize((50, 27))
        self.text_ctrl_loiter_radius.SetMinSize((50, 27))
        self.text_ctrl_wp_default_alt.SetMinSize((70, 27))
        self.label_home_lat_value.SetMinSize((100, 17))
        self.label_home_lat_value.SetForegroundColour(wx.Colour(0, 127, 255))
        self.label_home_lon_value.SetMinSize((100, 17))
        self.label_home_lon_value.SetForegroundColour(wx.Colour(0, 127, 255))
        self.label_home_alt_value.SetForegroundColour(wx.Colour(0, 127, 255))
        self.grid_mission.CreateGrid(0, 14)
        self.grid_mission.SetRowLabelSize(20)
        self.grid_mission.SetColLabelSize(20)
        self.grid_mission.SetColLabelValue(0, "Command")
        self.grid_mission.SetColSize(0, 150)
        self.grid_mission.SetColLabelValue(1, "P1")
        self.grid_mission.SetColLabelValue(2, "P2")
        self.grid_mission.SetColLabelValue(3, "P3")
        self.grid_mission.SetColLabelValue(4, "P4")
        self.grid_mission.SetColLabelValue(5, "Lat")
        self.grid_mission.SetColLabelValue(6, "Lon")
        self.grid_mission.SetColLabelValue(7, "Alt")
        self.grid_mission.SetColLabelValue(8, "Frame")
        self.grid_mission.SetColLabelValue(9, "Delete")
        self.grid_mission.SetColLabelValue(10, "Up")
        self.grid_mission.SetColLabelValue(11, "Down")
        self.grid_mission.SetDefaultColSize(-1)
        self.grid_mission.SetColLabelValue(12, "Distance")
        self.grid_mission.SetColLabelValue(13, "Grad (deg)")
        # end wxGlade

    def __do_layout(self):
        # begin wxGlade: MissionEditorFrame.__do_layout
        sizer_3 = wx.BoxSizer(wx.VERTICAL)
        sizer_16 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_4 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_12 = wx.BoxSizer(wx.VERTICAL)
        sizer_14 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_13 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_1 = wx.BoxSizer(wx.VERTICAL)
        sizer_11 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_10 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_9 = wx.BoxSizer(wx.VERTICAL)
        sizer_5 = wx.BoxSizer(wx.VERTICAL)
        sizer_15 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_4.Add(self.label_sync_state, 0, wx.LEFT | wx.ALIGN_CENTER_VERTICAL, 10)
        sizer_4.Add((10, 0), 0, 0, 0)
        sizer_5.Add(self.label_wp_radius, 0, 0, 0)
        sizer_5.Add(self.text_ctrl_wp_radius, 0, 0, 0)
        sizer_5.Add(self.label_loiter_rad, 0, 0, 0)
        sizer_15.Add(self.text_ctrl_loiter_radius, 0, 0, 0)
        sizer_15.Add(self.checkbox_loiter_dir, 0, 0, 0)
        sizer_5.Add(sizer_15, 1, wx.EXPAND, 0)
        sizer_4.Add(sizer_5, 0, wx.EXPAND, 0)
        sizer_4.Add((20, 20), 0, 0, 0)
        sizer_9.Add(self.label_default_alt, 0, 0, 0)
        sizer_9.Add(self.text_ctrl_wp_default_alt, 0, 0, 0)
        sizer_4.Add(sizer_9, 0, wx.EXPAND, 0)
        sizer_4.Add((20, 20), 0, 0, 0)
        sizer_1.Add(self.label_home_location, 0, 0, 0)
        sizer_2.Add(self.label_home_lat, 0, 0, 0)
        sizer_2.Add((40, 0), 0, 0, 0)
        sizer_2.Add(self.label_home_lat_value, 0, 0, 0)
        sizer_1.Add(sizer_2, 1, wx.EXPAND, 0)
        sizer_10.Add(self.label_home_lon, 0, 0, 0)
        sizer_10.Add((36, 0), 0, 0, 0)
        sizer_10.Add(self.label_home_lon_value, 0, 0, 0)
        sizer_1.Add(sizer_10, 1, wx.EXPAND, 0)
        sizer_11.Add(self.label_home_alt, 0, 0, 0)
        sizer_11.Add((11, 0), 0, 0, 0)
        sizer_11.Add(self.label_home_alt_value, 0, 0, 0)
        sizer_1.Add(sizer_11, 1, wx.EXPAND, 0)
        sizer_4.Add(sizer_1, 0, wx.EXPAND, 0)
        sizer_13.Add(self.button_read_wps, 0, 0, 0)
        sizer_13.Add((20, 20), 0, 0, 0)
        sizer_13.Add(self.button_write_wps, 0, 0, 0)
        sizer_12.Add(sizer_13, 1, wx.EXPAND, 0)
        sizer_12.Add((20, 20), 0, 0, 0)
        sizer_14.Add(self.button_load_wp_file, 0, 0, 0)
        sizer_14.Add((20, 20), 0, 0, 0)
        sizer_14.Add(self.button_save_wp_file, 0, 0, 0)
        sizer_12.Add(sizer_14, 1, wx.EXPAND, 0)
        sizer_4.Add(sizer_12, 0, wx.EXPAND, 0)
        sizer_3.Add(sizer_4, 0, wx.EXPAND, 0)
        sizer_3.Add(self.grid_mission, 1, wx.EXPAND, 0)
        sizer_16.Add(self.button_add_wp, 0, 0, 0)
        sizer_16.Add(self.button_split, 0, 0, 0)
        sizer_3.Add(sizer_16, 0, wx.EXPAND, 0)
        self.SetSizer(sizer_3)
        self.Layout()
        # end wxGlade

    def set_event_queue(self, q):
        self.event_queue = q

    def set_event_queue_lock(self, l):
        self.event_queue_lock = l

    def set_gui_event_queue(self, q):
        self.gui_event_queue = q

    def set_gui_event_queue_lock(self, l):
        self.gui_event_queue_lock = l

    def set_close_window_semaphore(self, sem):
        self.close_window_semaphore = sem

    def time_to_process_gui_events(self, evt):
        event_processed = False
        queue_access_start_time = time.time()
        self.gui_event_queue_lock.acquire()
        while (not self.gui_event_queue.empty()) and (time.time() < queue_access_start_time) < 0.6:
            event_processed = True
            event = self.gui_event_queue.get()
            try:
                self.process_gui_event(event)
            except Exception as e:
                print("Caught exception (%s)" % str(e))

        self.gui_event_queue_lock.release()

        if (event_processed == True):
            #redraw window to apply changes
            self.Refresh()
            self.Update()

    def process_gui_event(self, event):
        if event.get_type() == me_event.MEGE_CLEAR_MISS_TABLE:
            self.grid_mission.ClearGrid()
            if (self.grid_mission.GetNumberRows() > 0):
                self.grid_mission.DeleteRows(0,
                        self.grid_mission.GetNumberRows())
            self.grid_mission.SetDefaultColSize(50, True)
            self.grid_mission.SetColSize(ME_COMMAND_COL, 150)
            self.grid_mission.SetColSize(ME_LAT_COL, 100)
            self.grid_mission.SetColSize(ME_LON_COL, 100)
            self.grid_mission.SetColSize(ME_ALT_COL, 75)
            self.grid_mission.SetColSize(ME_DIST_COL, 1)
            self.grid_mission.SetColSize(ME_ANGLE_COL, 1)

            self.grid_mission.ForceRefresh()
        elif event.get_type() == me_event.MEGE_ADD_MISS_TABLE_ROWS:
            num_new_rows = event.get_arg("num_rows")
            if (num_new_rows < 1):
                return
            old_num_rows = self.grid_mission.GetNumberRows()
            self.grid_mission.AppendRows(num_new_rows)
            self.prep_new_rows(old_num_rows, num_new_rows)
            self.grid_mission.ForceRefresh()
        elif event.get_type() == me_event.MEGE_SET_MISS_ITEM:
            row = event.get_arg("num") - 1
            command = event.get_arg("command")

            if row == -1:
                #1st mission item is special: it's the immutable home position
                self.label_home_lat_value.SetLabel(
                        str(event.get_arg("lat")))
                self.label_home_lon_value.SetLabel(
                        str(event.get_arg("lon")))
                self.label_home_alt_value.SetLabel(
                        str(event.get_arg("alt")))

            else: #not the first mission item
                if command in me_defines.miss_cmds:
                    self.grid_mission.SetCellValue(row, ME_COMMAND_COL,
                            me_defines.miss_cmds[command])
                else:
                    self.grid_mission.SetCellValue(row, ME_COMMAND_COL,
                            str(command))

                self.grid_mission.SetCellValue(row, ME_P1_COL,
                        str(event.get_arg("param1")))
                self.grid_mission.SetCellValue(row, ME_P2_COL,
                        str(event.get_arg("param2")))
                self.grid_mission.SetCellValue(row, ME_P3_COL,
                        str(event.get_arg("param3")))
                self.grid_mission.SetCellValue(row, ME_P4_COL,
                        str(event.get_arg("param4")))
                self.grid_mission.SetCellValue(row, ME_LAT_COL,
                        str(event.get_arg("lat")))
                self.grid_mission.SetCellValue(row, ME_LON_COL,
                        str(event.get_arg("lon")))
                self.grid_mission.SetCellValue(row, ME_ALT_COL,
                        "%.2f" % event.get_arg("alt"))
                self.set_grad_dist()

                frame_num = event.get_arg("frame")
                if frame_num in me_defines.frame_enum:
                    self.grid_mission.SetCellValue(row, ME_FRAME_COL,
                        me_defines.frame_enum[frame_num])
                else:
                    self.grid_mission.SetCellValue(row, ME_FRAME_COL, "Und")


        elif event.get_type() == me_event.MEGE_SET_WP_RAD:
            self.text_ctrl_wp_radius.SetValue(str(event.get_arg("wp_rad")))
            self.text_ctrl_wp_radius.SetForegroundColour(wx.Colour(0, 0, 0))
        elif event.get_type() == me_event.MEGE_SET_LOIT_RAD:
            loiter_radius = event.get_arg("loit_rad")
            self.text_ctrl_loiter_radius.SetValue(
                    str(math.fabs(loiter_radius)))
            self.text_ctrl_loiter_radius.SetForegroundColour(wx.Colour(0, 0, 0))
            if (loiter_radius < 0.0):
                self.checkbox_loiter_dir.SetValue(False)
            else:
                self.checkbox_loiter_dir.SetValue(True)
        elif event.get_type() == me_event.MEGE_SET_WP_DEFAULT_ALT:
            self.text_ctrl_wp_default_alt.SetValue(str(
                event.get_arg("def_wp_alt")))
            self.text_ctrl_wp_default_alt.SetForegroundColour(wx.Colour(0, 0, 0))
        elif event.get_type() == me_event.MEGE_SET_LAST_MAP_CLICK_POS:
            self.last_map_click_pos = event.get_arg("click_pos")

    def prep_new_row(self, row_num):
        command_choices = sorted(list(me_defines.miss_cmds.values()))

        cell_ed = wx.grid.GridCellChoiceEditor(command_choices)
        self.grid_mission.SetCellEditor(row_num, ME_COMMAND_COL, cell_ed)
        self.grid_mission.SetCellValue(row_num, ME_COMMAND_COL, "NAV_WAYPOINT")

        for i in range(1, 7):
            self.grid_mission.SetCellValue(row_num, i, "0.0")
        for i in range(12, 14):
            self.grid_mission.SetCellValue(row_num, i, "0.0")

        #set altitude to default:
        self.grid_mission.SetCellValue(row_num, ME_ALT_COL,
                self.text_ctrl_wp_default_alt.GetValue())

        #populate frm cell editor and set to default value

        frame_cell_ed = wx.grid.GridCellChoiceEditor(list(me_defines.frame_enum.values()))
        self.grid_mission.SetCellEditor(row_num, ME_FRAME_COL, frame_cell_ed)

        # default to previous rows frame
        if row_num > 0:
            frame = self.grid_mission.GetCellValue(row_num-1, ME_FRAME_COL)
            if frame not in ["Rel", "AGL"]:
                frame = "Rel"
        else:
            frame = "Rel"
        self.grid_mission.SetCellValue(row_num, ME_FRAME_COL, frame)

        #this makes newest row always have the cursor in it,
        #making the "Add Below" button work like I want:
        self.grid_mission.SetGridCursor(row_num, ME_COMMAND_COL)

    def prep_new_rows(self, start_row, num_rows):
        num_remaining = num_rows
        current_row = start_row
        while num_remaining > 0:
            self.prep_new_row(current_row)
            current_row = current_row + 1
            num_remaining = num_remaining - 1

    def set_modified_state(self, modified):
        if (modified):
            self.set_grad_dist()
            self.label_sync_state.SetLabel("MODIFIED")
            self.label_sync_state.SetForegroundColour(wx.Colour(255, 0, 0))
        else:
            self.label_sync_state.SetLabel("SYNCED")
            self.label_sync_state.SetForegroundColour(wx.Colour(12, 152, 26))

    def read_wp_pushed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_READ_WPS))

        #sneak in some queries about a few other items as well:
        self.event_queue.put(MissionEditorEvent(me_event.MEE_GET_WP_RAD))
        self.event_queue.put(MissionEditorEvent(me_event.MEE_GET_LOIT_RAD))
        self.event_queue.put(MissionEditorEvent(me_event.MEE_GET_WP_DEFAULT_ALT))

        self.event_queue_lock.release()
        event.Skip()

        #TODO: the read actually has to succeed before I can say this:
        self.set_modified_state(False)

    def write_wp_pushed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_WRITE_WPS,count=
            self.grid_mission.GetNumberRows()+1))

        #home point first:
        lat = float(self.label_home_lat_value.GetLabel())
        lon = float(self.label_home_lon_value.GetLabel())
        alt = float(self.label_home_alt_value.GetLabel())
        self.event_queue.put(MissionEditorEvent(me_event.MEE_WRITE_WP_NUM,
                                                num=0,cmd_id=16,p1=0.0,p2=0.0,p3=0.0,p4=0.0,
                                                lat=lat,lon=lon,alt=alt,frame=0))

        for i in range(0, self.grid_mission.GetNumberRows()):
            cmd_id = me_defines.cmd_reverse_lookup(self.grid_mission.GetCellValue(i,0))
            #don't lock up the misseditor on missing input!
            #anything missing will just be zero
            try:
                p1 = float(self.grid_mission.GetCellValue(i,ME_P1_COL))
            except:
                p1 = 0.0
            try:
                p2 = float(self.grid_mission.GetCellValue(i,ME_P2_COL))
            except:
                p2 = 0.0
            try:
                p3 = float(self.grid_mission.GetCellValue(i,ME_P3_COL))
            except:
                p3 = 0.0
            try:
                p4 = float(self.grid_mission.GetCellValue(i,ME_P4_COL))
            except:
                p4 = 0.0
            try:
                lat = float(self.grid_mission.GetCellValue(i,ME_LAT_COL))
            except:
                lat = 0.0
            try:
                lon = float(self.grid_mission.GetCellValue(i,ME_LON_COL))
            except:
                lon = 0.0
            try:
                alt = float(self.grid_mission.GetCellValue(i,ME_ALT_COL))
            except:
                alt = 0.0
            try:
                frame = float(me_defines.frame_enum_rev[self.grid_mission.GetCellValue(i,ME_FRAME_COL)])
            except:
                frame = 0.0

            self.event_queue.put(MissionEditorEvent(me_event.MEE_WRITE_WP_NUM,
                                                    num=i+1,cmd_id=cmd_id,p1=p1,p2=p2,p3=p3,p4=p4,
                                                    lat=lat,lon=lon,alt=alt,frame=frame))

        self.event_queue_lock.release()

        self.set_modified_state(False)

        event.Skip()

    def load_wp_file_pushed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        fd = wx.FileDialog(self, "Open Mission File", os.getcwd(), "",
                "MissionFiles(*.txt.*.wp,*.waypoints)|*.txt;*.wp;*.waypoints", wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        if (fd.ShowModal() == wx.ID_CANCEL):
            return #user changed their mind...

        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_LOAD_WP_FILE,
            path=fd.GetPath()))
        self.event_queue_lock.release()

        self.last_mission_file_path = fd.GetPath()

        event.Skip()

    def add_wp_below_pushed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        row_selected = self.grid_mission.GetGridCursorRow()
        if (row_selected < 0):
            row_selected = self.grid_mission.GetNumberRows()-1

        self.grid_mission.InsertRows(row_selected+1)
        self.prep_new_row(row_selected+1)

        #set lat/lon based on last map click position:
        if self.last_map_click_pos is not None:
            self.grid_mission.SetCellValue(row_selected + 1, ME_LAT_COL,
                    str(self.last_map_click_pos[0]))
            self.grid_mission.SetCellValue(row_selected + 1, ME_LON_COL,
                    str(self.last_map_click_pos[1]))
        #highlight new row
        self.grid_mission.SelectRow(row_selected+1)
        self.set_modified_state(True)

        self.fix_jumps(row_selected+1, 1)

        event.Skip()

    def split_pushed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        row_selected = self.grid_mission.GetGridCursorRow()
        if (row_selected < 2):
            print("Invalid row selected")
            event.Skip()
            return

        if self.grid_mission.GetCellValue(row_selected, ME_COMMAND_COL) != "NAV_WAYPOINT":
            print("Bad command (need NAV_WAYPOINT)")
            event.Skip()
            return

        if self.grid_mission.GetCellValue(row_selected-1, ME_COMMAND_COL) != "NAV_WAYPOINT":
            print("Bad previous command (need NAV_WAYPOINT)")
            event.Skip()
            return

        if (self.grid_mission.GetCellValue(row_selected, ME_FRAME_COL) !=
            self.grid_mission.GetCellValue(row_selected-1, ME_FRAME_COL)):
            print("Items differ in frame")
            event.Skip()
            return

        lat = (float(self.grid_mission.GetCellValue(row_selected, ME_LAT_COL)) +
               float(self.grid_mission.GetCellValue(row_selected-1, ME_LAT_COL)))/2
        lng = (float(self.grid_mission.GetCellValue(row_selected, ME_LON_COL)) +
               float(self.grid_mission.GetCellValue(row_selected-1, ME_LON_COL)))/2
        alt = (float(self.grid_mission.GetCellValue(row_selected, ME_ALT_COL)) +
               float(self.grid_mission.GetCellValue(row_selected-1, ME_ALT_COL)))/2
        self.grid_mission.InsertRows(row_selected)
        self.prep_new_row(row_selected)

        self.grid_mission.SetCellValue(row_selected, ME_LAT_COL, str(lat))
        self.grid_mission.SetCellValue(row_selected, ME_LON_COL, str(lng))
        self.grid_mission.SetCellValue(row_selected, ME_ALT_COL, str(alt))
        #highlight new row
        self.grid_mission.SelectRow(row_selected)
        self.set_modified_state(True)

        self.fix_jumps(row_selected, 1)

        event.Skip()

    def save_wp_file_pushed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        fd = wx.FileDialog(self, "Save Mission File", os.getcwd(),
                           os.path.basename(self.last_mission_file_path), "MissionFiles(*.txt.*.wp,*.waypoints)|*.txt;*.wp;*.waypoints",
                               wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if (fd.ShowModal() == wx.ID_CANCEL):
            return #user change their mind...

        #ask mp_misseditor module to save file
        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_SAVE_WP_FILE,
            path=fd.GetPath()))
        self.event_queue_lock.release()

        self.last_mission_file_path = fd.GetPath()

        event.Skip()

    def on_mission_grid_cell_select(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        self.grid_mission.SetColLabelValue(ME_P1_COL, "P1")
        self.grid_mission.SetColLabelValue(ME_P2_COL, "P2")
        self.grid_mission.SetColLabelValue(ME_P3_COL, "P3")
        self.grid_mission.SetColLabelValue(ME_P4_COL, "P4")
        self.grid_mission.SetColLabelValue(ME_LAT_COL, "Lat")
        self.grid_mission.SetColLabelValue(ME_LON_COL, "Lon")
        self.grid_mission.SetColLabelValue(ME_ALT_COL, "Alt")
        self.grid_mission.SetColLabelValue(ME_DIST_COL, "Distance")
        self.grid_mission.SetColLabelValue(ME_ANGLE_COL, "Grad (deg)")

        if event.GetRow() == self.grid_mission.GetNumberRows():
            # this event fires when last row is deleted; ignore
            # attempt to set a cell in that deleted row as selected.
            return
        command = self.grid_mission.GetCellValue(event.GetRow(), ME_COMMAND_COL)

        col_labels = me_defines.get_column_labels(command)
        for col in col_labels.keys():
            self.grid_mission.SetColLabelValue(col, col_labels[col])

        event.Skip()

    def on_mission_grid_cell_left_click(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        #delete column?
        if (event.GetCol() == ME_DELETE_COL):
            row = event.GetRow()
            dlg = wx.MessageDialog(self, 'Sure you want to delete item ' + str(row+1) + '?', 'Really Delete?', wx.YES_NO | wx.ICON_EXCLAMATION)
            result = dlg.ShowModal()

            if (result == wx.ID_YES):
                #delete this row
                self.grid_mission.DeleteRows(row)
                self.set_modified_state(True)
                self.fix_jumps(row, -1)
        #up column?
        elif (event.GetCol() == ME_UP_COL):
            row = event.GetRow()
            if (row == 0): #can't go any higher
                return
            #insert a copy of this row above the previous row, then delete this row
            self.grid_mission.InsertRows(row-1)
            self.prep_new_row(row-1)
            for i in range(0, self.grid_mission.GetNumberCols()):
                self.grid_mission.SetCellValue(row-1, i,
                    self.grid_mission.GetCellValue(row+1,i))
            self.grid_mission.DeleteRows(row+1)
            #move the cursor to where the row moved and highlight the row
            self.grid_mission.SetGridCursor(row-1,ME_UP_COL)
            self.grid_mission.SelectRow(row-1)

            self.set_modified_state(True)

            #Return so we don't skip the event so the cursor will go where expected
            return

        #down column?
        elif (event.GetCol() == ME_DOWN_COL):
            row = event.GetRow()
            if (row == self.grid_mission.GetNumberRows() - 1): #can't go lower
                return

            #insert a copy of this row 2 rows down, then delete this row
            self.grid_mission.InsertRows(row+2)
            self.prep_new_row(row+2)
            for i in range(0, self.grid_mission.GetNumberCols()):
                self.grid_mission.SetCellValue(row+2, i,
                    self.grid_mission.GetCellValue(row, i))
            self.grid_mission.DeleteRows(row)

            #move the cursor to where the row moved and highlight the row
            self.grid_mission.SetGridCursor(row+1,ME_DOWN_COL)
            self.grid_mission.SelectRow(row+1)

            self.set_modified_state(True)

            #Return so we don't skip the event so the cursor will go where expected
            return

        event.Skip()

    def on_mission_grid_cell_changed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        self.set_modified_state(True)
        event.Skip()
    def on_wp_radius_changed(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        #change text red
        self.text_ctrl_wp_radius.SetForegroundColour(wx.Colour(255, 0, 0))
        event.Skip()
    def on_wp_radius_enter(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        try:
            wp_radius = float(self.text_ctrl_wp_radius.GetValue())
        except Exception as e:
            print(str(e))
            return

        #text back to black
        self.text_ctrl_wp_radius.SetForegroundColour(wx.Colour(0, 0, 0))

        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_SET_WP_RAD,
            rad=wp_radius))
        self.event_queue_lock.release()

        event.Skip()

    def send_loiter_rad_msg(self):
        try:
            loit_radius = float(self.text_ctrl_loiter_radius.GetValue())
        except Exception as e:
            print(str(e))
            return

        if (not self.checkbox_loiter_dir.IsChecked()):
            loit_radius = loit_radius * -1.0

        #text back to black
        self.text_ctrl_loiter_radius.SetForegroundColour(wx.Colour(0, 0, 0))

        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_SET_LOIT_RAD,
            rad=loit_radius))
        self.event_queue_lock.release()

    def on_loiter_rad_enter(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        self.send_loiter_rad_msg()

        event.Skip()

    def on_loiter_rad_change(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        #change text red
        self.text_ctrl_loiter_radius.SetForegroundColour(wx.Colour(255, 0, 0))

        event.Skip()
    def on_loiter_dir_cb_change(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        self.send_loiter_rad_msg()
        event.Skip()

    def on_wp_default_alt_enter(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        try:
            def_alt = float(self.text_ctrl_wp_default_alt.GetValue())
        except Exception as e:
            print(str(e))
            return

        #text back to black
        self.text_ctrl_wp_default_alt.SetForegroundColour(wx.Colour(0, 0, 0))

        self.event_queue_lock.acquire()
        self.event_queue.put(MissionEditorEvent(me_event.MEE_SET_WP_DEFAULT_ALT,
            alt=def_alt))
        self.event_queue_lock.release()

        event.Skip()

    def on_wp_default_alt_change(self, event):  # wxGlade: MissionEditorFrame.<event_handler>
        #change text red
        self.text_ctrl_wp_default_alt.SetForegroundColour(wx.Colour(255, 0, 0))

        event.Skip()

    def fix_jumps(self, row_selected, delta):
        '''fix up jumps when we add/remove rows'''
        numrows = self.grid_mission.GetNumberRows()
        for row in range(numrows):
            command = self.grid_mission.GetCellValue(row, ME_COMMAND_COL)
            if command in ["DO_JUMP", "DO_CONDITION_JUMP"]:
                p1 = int(float(self.grid_mission.GetCellValue(row, ME_P1_COL)))
                if p1 > row_selected and p1+delta>0:
                    self.grid_mission.SetCellValue(row, ME_P1_COL, str(float(p1+delta)))

    def set_grad_dist(self):
        '''fix up distance and gradient when changing cell values'''
        home_def_alt = float(self.label_home_alt_value.GetLabel())
        numrows = self.grid_mission.GetNumberRows()
        for row in range(1,numrows):
            command = self.grid_mission.GetCellValue(row, ME_COMMAND_COL)
            command_prev = self.grid_mission.GetCellValue(row - 1, ME_COMMAND_COL)
            lat = float(self.grid_mission.GetCellValue(row, ME_LAT_COL))
            lon = float(self.grid_mission.GetCellValue(row, ME_LON_COL))
            if (lat == 0) and (lon == 0):
                dist = 0.0
                grad = 0.0
            else :
              if "NAV" in command:
                row_prev = row - 1
                while ("NAV" not in command_prev) and (row_prev > 0):
                    command_prev = self.grid_mission.GetCellValue(row_prev - 1, ME_COMMAND_COL)
                    row_prev = row_prev - 1
                prev_lat = float(self.grid_mission.GetCellValue(row_prev, ME_LAT_COL))
                prev_lon = float(self.grid_mission.GetCellValue(row_prev, ME_LON_COL))
                prev_alt = float(self.grid_mission.GetCellValue(row_prev, ME_ALT_COL))
                if (self.grid_mission.GetCellValue(row_prev, ME_FRAME_COL) == "Rel"):
                    prev_alt = prev_alt + home_def_alt
                elif (self.grid_mission.GetCellValue(row_prev, ME_FRAME_COL) == "AGL"):
                    prev_alt = self.ElevationModel.GetElevation(prev_lat, prev_lon) + prev_alt
                while (prev_lat == 0) and (prev_lon == 0) and (row_prev > 0):
                    prev_lat = float(self.grid_mission.GetCellValue(row_prev - 1, ME_LAT_COL))
                    prev_lon = float(self.grid_mission.GetCellValue(row_prev - 1, ME_LON_COL))
                    row_prev = row_prev - 1
                dist = mp_util.gps_distance(lat, lon, prev_lat, prev_lon)
                curr_alt = float(self.grid_mission.GetCellValue(row, ME_ALT_COL))
                if (self.grid_mission.GetCellValue(row, ME_FRAME_COL) == "Rel"):
                    curr_alt = curr_alt + home_def_alt
                elif(self.grid_mission.GetCellValue(row, ME_FRAME_COL) == "AGL"):
                    curr_alt = curr_alt + self.ElevationModel.GetElevation(lat, lon)
                grad = math.atan2(curr_alt - prev_alt, dist) *180 / math.pi
              else:
                grad = 0.0
                dist = 0.0
            dist = format(dist, '.1f')
            grad = format(grad, '.1f')
            self.grid_mission.SetCellValue(row, ME_DIST_COL, str(dist))
            self.grid_mission.SetCellValue(row, ME_ANGLE_COL, str(grad))

    def on_idle(self, event):
        now = time.time()
        if now - self.last_layout_send > 1:
            self.last_layout_send = now
            self.event_queue.put(win_layout.get_wx_window_layout(self))

        obj = None
        while not self.state.object_queue.empty():
            obj = self.state.object_queue.get()
            if isinstance(obj, win_layout.WinLayout):
                win_layout.set_wx_window_layout(self, obj)

# end of class MissionEditorFrame
if __name__ == "__main__":
    app = wx.App(False)
    wx.InitAllImageHandlers()
    missionEditor = (None, wx.ID_ANY, "")
    app.SetTopWindow(missionEditor)
    missionEditor.Show()
    app.MainLoop()



#!/usr/bin/env python
'''monitor sensor consistancy'''

import time, math
from pymavlink import mavutil

from MAVProxy.modules.lib import mp_module


def angle_diff(angle1, angle2):
    ret = angle1 - angle2
    if ret > 180:
        ret -= 360;
    if ret < -180:
        ret += 360
    return ret

class sensors_report(object):
    def __init__(self):
        self.last_report = 0
        self.ok = True
        self.value = 0

class SensorsModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(SensorsModule, self).__init__(mpstate, "sensors", "monitor sensor consistancy")
        self.add_command('sensors', self.cmd_sensors, "show key sensors")
        self.add_command('speed', self.cmd_speed, "enable/disable speed report")

        self.last_report = 0
        self.ok = True
        self.value = 0
        self.ground_alt = 0
        self.gps_alt = 0
        self.max_speed = 0
        self.last_watch = 0
        self.reports = {}
        self.reports['heading'] = sensors_report()
        self.reports['altitude'] = sensors_report()
        self.reports['speed'] = sensors_report()

        from MAVProxy.modules.lib.mp_settings import MPSetting
        self.settings.append(MPSetting('speedreporting', bool, False, 'Speed Reporting', tab='Sensors'))

        if 'GPS_RAW_INT' in self.status.msgs:
            # cope with reload
            gps = mpstate.status.msgs['GPS_RAW_INT']
            self.ground_alt = (gps.alt / 1.0e3) - self.status.altitude

    def cmd_sensors(self, args):
        '''show key sensors'''
        gps_heading = self.status.msgs['GPS_RAW_INT'].cog * 0.01

        self.console.writeln("heading: %u/%u   alt: %u/%u  r/p: %u/%u speed: %u/%u  thr: %u" % (
            self.status.msgs['VFR_HUD'].heading,
            gps_heading,
            self.status.altitude,
            self.gps_alt,
            math.degrees(self.status.msgs['ATTITUDE'].roll),
            math.degrees(self.status.msgs['ATTITUDE'].pitch),
            self.status.msgs['VFR_HUD'].airspeed,
            self.status.msgs['VFR_HUD'].groundspeed,
            self.status.msgs['VFR_HUD'].throttle))


    def cmd_speed(self, args):
        '''enable/disable speed report'''
        self.settings.set('speedreporting', not self.settings.speedreporting)
        if self.settings.speedreporting:
            self.console.writeln("Speed reporting enabled", bg='yellow')
        else:
            self.console.writeln("Speed reporting disabled", bg='yellow')

    def report(self, name, ok, msg=None, deltat=20):
        '''report a sensor error'''
        r = self.reports[name]
        if time.time() < r.last_report + deltat:
            r.ok = ok
            return
        r.last_report = time.time()
        if ok and not r.ok:
            self.say("%s OK" % name)
        r.ok = ok
        if not r.ok:
            self.say(msg)

    def report_change(self, name, value, maxdiff=1, deltat=10):
        '''report a sensor change'''
        r = self.reports[name]
        if time.time() < r.last_report + deltat:
            return
        r.last_report = time.time()
        if math.fabs(r.value - value) < maxdiff:
            return
        r.value = value
        self.say("%s %u" % (name, value))

    def check_heading(self, m):
        '''check heading discrepancy'''
        if 'GPS_RAW_INT' in self.status.msgs:
            gps = self.status.msgs['GPS_RAW_INT']
            if gps.vel < 300:
                return
            diff = math.fabs(angle_diff(m.heading, gps.cog / 100.0))
        else:
            return
        self.report('heading', diff < 20, 'heading error %u' % diff)

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == 'VFR_HUD' and 'GPS_RAW_INT' in self.status.msgs:
            self.check_heading(m)
            if self.settings.speedreporting:
                if m.airspeed != 0:
                    speed = m.airspeed
                else:
                    speed = m.groundspeed
                speed = self.speed_convert_units(speed)
                self.report_change('speed', speed, maxdiff=2, deltat=2)
        if self.status.watch == "sensors" and time.time() > self.sensors_state.last_watch + 1:
            self.sensors_state.last_watch = time.time()
            self.cmd_sensors([])

def init(mpstate):
    '''initialise module'''
    return SensorsModule(mpstate)



'''Advise user if there are applicable user alerts
This is a reference implementation based on the documentation at
https://github.com/ArduPilot/useralerts and
https://ardupilot.org/dev/docs/user-alerts-developer.html '''


import sys
import json
import string
from packaging import version
from pymavlink import mavutil

if sys.version_info.major < 3:
    from urllib2 import Request
    from urllib2 import urlopen
    from urllib2 import URLError
else:
    from urllib.request import Request
    from urllib.request import urlopen
    from urllib.error import URLError

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib.mp_util import decode_flight_sw_version

class UserAlertsModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(UserAlertsModule, self).__init__(mpstate, "useralerts", public=True)

        self.url = "https://firmware.ardupilot.org/useralerts/manifest.json"
        self.testurl = "https://firmware.ardupilot.org/useralerts/examplemanifest.json"

        # The useTest setting allows the user to query the example database
        # instead of the real database
        # The filterBoard setting determines whether User Alerts are filtered
        # by the hardwareLimited field.
        self.alerts_settings = mp_settings.MPSettings(
            [('useTest', bool, False),
             ('filterBoard', bool, True)])
        self.add_completion_function('(ALERTSETTING)',
                                     self.alerts_settings.completion)

        self.add_command('useralerts', self.cmd_check, "Check User Alerts",
                         ["check",
                          "set (ALERTSETTING)"])

        self.board = None
        self.version = None

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        mtype = m.get_type()
        if mtype == "STATUSTEXT":
            #Looking for a response to MAV_CMD_DO_SEND_BANNER:
            #APM: KakuteF4 00390032 4B4B500F 20313548
            #no nice way to parse this. So assuming 4 words, the last 3 are 8 character hex values
            splitText = m.text.split(" ")
            if (len(splitText) == 4 and len(splitText[1]) == 8 and len(splitText[2]) == 8 and len(splitText[3]) == 8 and
                    all(c in string.hexdigits for c in splitText[1]) and all(c in string.hexdigits for c in splitText[2]) and
                    all(c in string.hexdigits for c in splitText[3])):
                self.board = splitText[0]
            if not self.alerts_settings.filterBoard:
                self.board = "N/A"

        elif mtype == "AUTOPILOT_VERSION":
            #Example output:
            #AUTOPILOT_VERSION {capabilities : 61935, flight_sw_version : 67174400, middleware_sw_version : 0, os_sw_version : 0,
            # board_version : 0, flight_custom_version : [54, 49, 51, 100, 50, 99, 50, 101],
            # middleware_custom_version : [0, 0, 0, 0, 0, 0, 0, 0], os_custom_version : [51, 51, 49, 102, 101, 55, 53, 100],
            # vendor_id : 0, product_id : 0, uid : 0, uid2 : [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]}
            vMajor,vMinor,vPatch,vFwType = decode_flight_sw_version(m.flight_sw_version)
            self.version = "{0}.{1}.{2}".format(vMajor, vMinor, vPatch)

        if self.board and self.version:
            print("Checking Alerts for {1} version {0}, board {2}".format(self.version, self.mpstate.vehicle_type, self.board))

            #download user alerts
            if self.alerts_settings.useTest:
                req = Request(self.testurl)
            else:
                req = Request(self.url)

            try:
                r = urlopen(req).read()
            except URLError:
                print("Error: unable to download User Alerts")
                # reset for next time "check" is called
                self.version = None
                self.board = None
                return

            allAlerts = json.loads(r.decode('utf-8'))
            print("Downloaded {0} alerts".format(len(allAlerts)))

            # Check and print out applicable alerts
            for (key, alert) in allAlerts.items():
                #print("Checking: {0}".format(key))
                (valid, mitigation) = self.alert_is_applicable(alert, self.version, str(self.mpstate.vehicle_type), self.board)
                if valid:
                    print("******************")
                    print("User alert {0} is valid for this vehicle:".format(key))
                    print("Issue: {0}\nMitigation: {1}".format(alert['description'], mitigation))

            # reset for next time "check" is called
            self.version = None
            self.board = None

    def alert_is_applicable(self, alert, APversion, APfirmware, boardName):
        '''checks for alert validity given ArduPilot version, board and type
        returns (true, mitigationString) if alert is applicable, (false, None) otherwise'''

        # if affectedFirmware is not all AND does not contain the APFirmware, alert is not applicable
        if ('all' not in alert['affectedFirmware']) and (APfirmware not in alert['affectedFirmware']):
            #print("False on affectedFirmware: " + str(alert['affectedFirmware']))
            return (False, None)

        # Check hardwareLimited. If there is a list of boards, check if on list
        if len(alert['hardwareLimited']) > 0 and (boardName not in alert['hardwareLimited']) and boardName != "N/A":
            #print("False on hardwareLimited: " + str(alert['hardwareLimited']))
            return (False, None)

        #check that versionFrom is either blank or less-than-or-equal-to the reported firmware
        if len(alert['versionFrom']) == 0 or version.parse(alert['versionFrom'][APfirmware]) <= version.parse(APversion):
            # Check if there is a fixed version out
            if APfirmware in alert['versionFixed'] and version.parse(alert['versionFixed'][APfirmware]) <= version.parse(APversion):
                # User is already running fixed version (or higher)
                #print("False on versionFixed: " + str(alert['versionFixed'][APfirmware]) + ", " + APversion)
                return (False, None)
            elif APfirmware in alert['versionFixed']:
                # There is a fixed version and the user is not running it
                return (True, "Upgrade to V{0} (or later) of {1}".format(alert['versionFixed'][APfirmware], APfirmware))
            else:
                # There is not a fixed version out yet for this firmware
                return (True, alert['mitigation'])
        else:
            #print("False on versionFrom: " + str(alert['versionFrom'][APfirmware]) + ", " + APversion)
            return (False, None)

    def cmd_doCheck(self):
        '''Query ArduPilot for version and board information
        When the messages are returned via mavlink_packet(), perform
        the checks for applicable User Alerts'''
        self.board = None
        self.version = None
        for messageID in [mavutil.mavlink.MAV_CMD_REQUEST_AUTOPILOT_CAPABILITIES, mavutil.mavlink.MAV_CMD_DO_SEND_BANNER]:
            self.master.mav.command_long_send(
                self.settings.target_system,  # target_system
                self.settings.target_component, # target_component
                messageID, # command
                1, # confirmation
                1, # param1
                0, # param2
                0, # param3
                0, # param4
                0, # param5
                0, # param6
                0) # param7
        #self.master.mav.command_long_send(
        #    self.settings.target_system,  # target_system
        #    self.settings.target_component, # target_component
        #    mavutil.mavlink.MAV_CMD_DO_SEND_BANNER, # command
        #    1, # confirmation
        #    0, # param1
        #    0, # param2
        #    0, # param3
        #    0, # param4
        #    0, # param5
        #    0, # param6
        #    0) # param7

    def cmd_check(self, args):
        '''Useralert operations'''
        usage = "Usage: useralerts <check|set>"
        if len(args) < 1:
            print(usage)
            return
        if args[0] == 'check':
            self.cmd_doCheck()
        elif args[0] == "set":
            self.alerts_settings.command(args[1:])

def init(mpstate):
    '''initialise module'''
    return UserAlertsModule(mpstate)



#!/usr/bin/env python

"""
  MAVProxy horizon indicator.
"""
from MAVProxy.modules.lib import multiproc
import time

class HorizonIndicator():
    '''
    A horizon indicator for MAVProxy.
    '''
    def __init__(self,title='MAVProxy: Horizon Indicator'):
        self.title  = title
        # Create Pipe to send attitude information from module to UI
        self.child_pipe_recv,self.parent_pipe_send = multiproc.Pipe()
        self.close_event = multiproc.Event()
        self.close_event.clear()
        self.child = multiproc.Process(target=self.child_task)
        self.child.start()
        self.child_pipe_recv.close()

    def child_task(self):
        '''child process - this holds all the GUI elements'''
        self.parent_pipe_send.close()
        
        from MAVProxy.modules.lib import wx_processguard
        from MAVProxy.modules.lib.wx_loader import wx
        from MAVProxy.modules.lib.wxhorizon_ui import HorizonFrame
        # Create wx application
        app = wx.App(False)
        app.frame = HorizonFrame(state=self, title=self.title)
        app.frame.SetDoubleBuffered(True)
        app.frame.Show()
        app.MainLoop()
        self.close_event.set()   # indicate that the GUI has closed

    def close(self):
        '''Close the window.'''
        self.close_event.set()
        if self.is_alive():
            self.child.join(2)

    def is_alive(self):
        '''check if child is still going'''
        return self.child.is_alive()

if __name__ == "__main__":
    # test the console
    multiproc.freeze_support()
    horizon = HorizonIndicator()
    while horizon.is_alive():
        print('test')
        time.sleep(0.5)



#!/usr/bin/env python
'''Decode RTCM v3 messages'''

RTCMv3_PREAMBLE = 0xD3
POLYCRC24 = 0x1864CFB

import struct

class RTCM3:
    def __init__(self, debug=False):
        self.crc_table = None
        self.debug = debug
        self.reset()

    def get_packet(self):
        '''return bytearray of last parsed packet'''
        return self.parsed_pkt

    def get_packet_ID(self):
        '''get get of packet, or None'''
        if self.parsed_pkt is None or len(self.parsed_pkt) < 8:
            return None
        id, = struct.unpack('>H', self.parsed_pkt[3:5])
        id >>= 4
        return id

    def reset(self):
        '''reset state'''
        self.pkt = bytearray()
        self.pkt_len = 0
        self.parsed_pkt = None

    def parse(self):
        '''parse packet'''
        parity = self.pkt[-3:]
        crc1 = parity[0] << 16 | parity[1] << 8 | parity[2]
        crc2 = self.crc24(self.pkt[:-3])
        if crc1 != crc2:
            if self.debug:
                print("crc fail len=%u" % len(self.pkt))
            # look for preamble
            idx = self.pkt[1:].find(bytearray([RTCMv3_PREAMBLE]))
            if idx >= 0:
                self.pkt = self.pkt[1+idx:]
                if len(self.pkt) >= 3:
                    self.pkt_len, = struct.unpack('>H', self.pkt[1:3])
                    self.pkt_len &= 0x3ff
                else:
                    self.pkt_len = 0
            else:
                self.reset()
            return False

        # got a good packet
        self.parsed_pkt = self.pkt
        self.pkt = bytearray()
        self.pkt_len = 0
        return True

    def read(self, byte):
        '''read in one byte, return true if a full packet is available'''

        #import random
        #if random.uniform(0,1000) < 1:
        #    return False

        byte = ord(byte)
        if len(self.pkt) == 0 and byte != RTCMv3_PREAMBLE:
            # discard
            return False

        self.pkt.append(byte)
        if self.pkt_len == 0 and len(self.pkt) >= 3:
            self.pkt_len, = struct.unpack('>H', self.pkt[1:3])
            self.pkt_len &= 0x3ff
            if self.pkt_len == 0:
                self.reset()
                return False

        if self.pkt_len > 0 and len(self.pkt) >= 3 + self.pkt_len + 3:
            remainder = self.pkt[6+self.pkt_len:]
            self.pkt = self.pkt[:6+self.pkt_len]
            # got header, packet body and parity
            ret = self.parse()
            self.pkt.extend(remainder)
            return ret

        # need more bytes
        return False

    def crc24(self, bytes):
        '''calculate 24 bit crc'''
        if self.crc_table is None:
            # initialise table
            self.crc_table = [0] * 256
            for i in range(256):
                self.crc_table[i] = i<<16
                for j in range(8):
                    self.crc_table[i] <<= 1
                    if (self.crc_table[i] & 0x1000000):
                        self.crc_table[i] ^= POLYCRC24
        crc = 0
        for b in bytes:
            crc = ((crc<<8)&0xFFFFFF) ^ self.crc_table[(crc>>16) ^ b]
        return crc

if __name__ == '__main__':
    from argparse import ArgumentParser
    import time
    parser = ArgumentParser(description='RTCM3 parser')

    parser.add_argument("filename", type=str, help="input file")
    parser.add_argument("--debug", action='store_true', help="show errors")
    parser.add_argument("--follow", action='store_true', help="continue reading on EOF")
    args = parser.parse_args()

    rtcm3 = RTCM3(args.debug)
    f = open(args.filename, 'rb')
    while True:
        b = f.read(1)
        if len(b) == 0:
            if args.follow:
                time.sleep(0.1)
                continue
            break
        if rtcm3.read(b):
            print("packet len %u ID %u" % (len(rtcm3.get_packet()), rtcm3.get_packet_ID()))



#!/usr/bin/env python

'''
Proximity sensor module
Peter Barker, August 2020

Deals with proximity sensor data like 360-degree lidars from the
vehicle.  Plots them on the map, for example.

'''

import os
import os.path
import sys
from pymavlink import mavutil
import errno
import time
import math

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import mp_settings

from MAVProxy.modules.mavproxy_map import mp_slipmap

class proximity(mp_module.MPModule):
    def __init__(self, mpstate, multi_vehicle=True):
        """Initialise module"""
        super(proximity, self).__init__(mpstate, "proximity", "")

        self.proximity_settings = mp_settings.MPSettings(
            [ ('verbose', bool, False),
          ])
        self.add_command('proximity',
                         self.cmd_proximity,
                         "proximity module",
                         ['set (PROXIMITYSETTING)',
                         ]
        )

        self.heading_for_vehicle = {}
        self.have_global_position = False

        # lat/lon per system ID
        self.lat_lon_for_vehicle = {}

    def usage(self):
        '''show help on command line options'''
        return "Usage: proximity <set>"

    def cmd_proximity(self, args):
        '''control behaviour of the module'''
        if len(args) == 0:
            print(self.usage())
        elif args[0] == "set":
            self.proximity_settings.command(args[1:])
        else:
            print(self.usage())

    def idle_task(self):
        '''called rapidly by mavproxy'''

    def foreach_map(self, closure):
        for mp in self.module_matching('map*'):
            closure(mp.map)

    def mavlink_packet_distance_sensor(self, vehicle, m):
        heading = self.heading_for_vehicle.get(vehicle, 0)
        tlayer = "Distance sensor for %u.%u id=%u" % (
            m.get_srcSystem(), m.get_srcComponent(), m.id)
        slipkey = '%s-POS%u' % (tlayer, m.orientation)
        if not m.get_srcSystem() in self.lat_lon_for_vehicle:
            return
        if m.current_distance == m.max_distance:
            self.foreach_map(lambda a_map : a_map.remove_object(slipkey))
            return

        (lat,lon) = self.lat_lon_for_vehicle[m.get_srcSystem()]
        mav_sensor_rotation_to_degrees = {
            mavutil.mavlink.MAV_SENSOR_ROTATION_NONE: 0,
            mavutil.mavlink.MAV_SENSOR_ROTATION_YAW_45: 45,
            mavutil.mavlink.MAV_SENSOR_ROTATION_YAW_90: 90,
            mavutil.mavlink.MAV_SENSOR_ROTATION_YAW_135 : 135,
            mavutil.mavlink.MAV_SENSOR_ROTATION_YAW_180: 180,
            mavutil.mavlink.MAV_SENSOR_ROTATION_YAW_225 : 225,
            mavutil.mavlink.MAV_SENSOR_ROTATION_YAW_270 : 270,
            mavutil.mavlink.MAV_SENSOR_ROTATION_YAW_315 : 315,
        }
        if m.orientation in mav_sensor_rotation_to_degrees:
            degrees = mav_sensor_rotation_to_degrees[m.orientation]
        else:
#            print("bad orientation (%u)" % m.orientation)
            return
        if m.current_distance >= m.max_distance:
            return

        p = mp_util.gps_newpos(lat, lon, heading+degrees, 0)
        # start angle/end angle are either side of the primary axis,
        # which is rotated to be North
        start_angle = -22.5
        end_angle = 22.5
        self.foreach_map(lambda a_map :
                         a_map.add_object(mp_slipmap.SlipCircle(
                             slipkey,
                             3,
                             p,
                             m.current_distance/100.0,
                             (255, 127, 0),
                             linewidth=3,
                             start_angle=start_angle,
                             end_angle=end_angle,
                             rotation=(-90+(heading+degrees))%360,
                         ))
        )

    def mavlink_packet_obstacle_distance(self, vehicle, m):
        heading = self.heading_for_vehicle.get(vehicle, 0)
        tlayer = "Obstacle Distance for %u.%u type=%u" % (
            m.get_srcSystem(), m.get_srcComponent(), m.sensor_type)
        if m.get_srcSystem() in self.lat_lon_for_vehicle:
            color = (255, 0, 255)
            (lat,lon) = self.lat_lon_for_vehicle[m.get_srcSystem()]
            self.foreach_map(lambda a_map : a_map.add_object(mp_slipmap.SlipClearLayer(tlayer)))
            increment = m.increment_f
            if increment == 0:
                increment = float(m.increment)
            measurement_count = 0
            for i in range(0, 72):
                if m.distances[i] != 65535:
                    measurement_count += 1
            fov = measurement_count*increment
            start_angle = -increment/2.0
            end_angle = increment/2.0
            rotation_start = -90 + heading
            for i in range(0, 72):
                slipkey = '%s-POS%u' % (tlayer, i)
                if m.distances[i] == m.max_distance+1:
                    # no measurement
                    self.foreach_map(lambda a_map : a_map.remove_object(slipkey))
                    measurement_count += 1
                    continue
                distance = m.distances[i] / 100.0 # cm -> m
                circle = mp_slipmap.SlipCircle(
                    slipkey,
                    3,
                    (lat, lon),
                    distance,
                    color,
                    linewidth=3,
                    start_angle=start_angle,
                    end_angle=end_angle,
                    rotation=(m.angle_offset+rotation_start+i*increment)%360,
                )
                if m.distances[i] < m.min_distance:
                    measurement_count += 1
                self.foreach_map(lambda a_map : a_map.add_object(circle))

            slipkey = "%s-range" % tlayer
            if fov == 360:
                # perfect circle
                # add a circle for max-range
                self.foreach_map(lambda a_map : a_map.add_object(mp_slipmap.SlipCircle(
                    slipkey,
                    3,
                    (lat, lon),
                    m.max_distance/100.0,
                    color,
                    linewidth=1,
                )))
            else:
                # add an arc for max-range
                self.foreach_map(lambda a_map : a_map.add_object(mp_slipmap.SlipCircle(
                    slipkey,
                    3,
                    (lat, lon),
                    m.max_distance/100.0,
                    color,
                    linewidth=1,
                    start_angle=-fov/2,
                    end_angle=fov/2,
                    rotation=(m.angle_offset+-90+(heading))%360,
                    add_radii = True,
                )))

    def mavlink_packet(self, m):
        '''handle mavlink packets'''
        mtype = m.get_type()
        vehicle = 'Vehicle%u' % m.get_srcSystem()
        if mtype == 'GLOBAL_POSITION_INT':
            (lat, lon, heading) = (m.lat*1.0e-7, m.lon*1.0e-7, m.hdg*0.01)
            self.heading_for_vehicle[vehicle] = heading
            self.lat_lon_for_vehicle[m.get_srcSystem()] = (lat,lon)
            if abs(lat) > 1.0e-3 or abs(lon) > 1.0e-3:
                self.have_global_position = True
        elif mtype == 'LOCAL_POSITION_NED' and not self.have_global_position:
            self.heading_for_vehicle[vehicle] = math.degrees(math.atan2(m.vy, m.vx))
        elif mtype == 'DISTANCE_SENSOR':
            self.mavlink_packet_distance_sensor(vehicle, m)
        elif mtype == 'OBSTACLE_DISTANCE':
            self.mavlink_packet_obstacle_distance(vehicle, m)

def init(mpstate):
    '''initialise module'''
    return proximity(mpstate)



'''
Support for ADS-B data
Samuel Dudley
Dec 2015
'''

from math import *

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from pymavlink import mavutil

obc_icons = {
    100 : 'greenplane.png',
    101 : 'greenplane.png',
    102 : 'cloud.png',
    103 : 'migbird.png',
    104 : 'hawk.png'
}

obc_radius = {
    100 : 300,
    101 : 300,
    102 : 173,
    103 : 100,
    104 : 200
}
    
def get_threat_radius(m):

    if m.emitter_type == 255:
        ''' objectAvoidance Database item, squawk contains radius in cm'''
        return m.squawk * 0.01
    '''get threat radius for an OBC item'''
    return obc_radius.get(m.emitter_type,0)


def get_threat_icon(m, default_icon):

    if m.emitter_type == 255:
        ''' objectAvoidance Database item, do not draw an icon'''
        return None
    '''get threat radius for an OBC item, else the true ADSB icon'''
    return obc_icons.get(m.emitter_type, default_icon)


class ADSBVehicle(object):
    '''a generic ADS-B threat'''

    def __init__(self, id, state):
        self.id = id
        self.state = state
        self.vehicle_colour = 'green'  # use plane icon for now
        self.vehicle_type = 'plane'
        self.icon = self.vehicle_colour + self.vehicle_type + '.png'
        self.update_time = 0
        self.is_evading_threat = False
        self.v_distance = None
        self.h_distance = None
        self.distance = None

    def update(self, state, tnow):
        '''update the threat state'''
        self.state = state
        self.update_time = tnow


class ADSBModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(ADSBModule, self).__init__(mpstate, "adsb", "ADS-B data support", public = True)
        self.threat_vehicles = {}
        self.active_threat_ids = []  # holds all threat ids the vehicle is evading

        self.add_command('adsb', self.cmd_ADSB, "adsb control",
                         ["<status>", "set (ADSBSETTING)"])

        self.ADSB_settings = mp_settings.MPSettings([("timeout", int, 5),  # seconds
                                                     ("threat_radius", int, 200),  # meters
                                                     ("show_threat_radius", bool, False),
                                                     # threat_radius_clear = threat_radius*threat_radius_clear_multiplier
                                                     ("threat_radius_clear_multiplier", int, 2),
                                                     ("show_threat_radius_clear", bool, False)])
        self.add_completion_function('(ADSBSETTING)',
                                     self.ADSB_settings.completion)
        
        self.threat_detection_timer = mavutil.periodic_event(2)
        self.threat_timeout_timer = mavutil.periodic_event(2)
        self.tnow = self.get_time()

    def cmd_ADSB(self, args):
        '''adsb command parser'''
        usage = "usage: adsb <set>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "status":
            print("total threat count: %u  active threat count: %u" %
                  (len(self.threat_vehicles), len(self.active_threat_ids)))

            for id in self.threat_vehicles.keys():
                print("id: %s  distance: %.2f m callsign: %s  alt: %.2f" % (id,
                                                                            self.threat_vehicles[id].distance,
                                                                            self.threat_vehicles[id].state['callsign'],
                                                                            self.threat_vehicles[id].state['altitude']))
        elif args[0] == "set":
            self.ADSB_settings.command(args[1:])
        else:
            print(usage)

    def perform_threat_detection(self):
        '''determine threats'''
        # TODO: perform more advanced threat detection
        threat_radius_clear = self.ADSB_settings.threat_radius * \
            self.ADSB_settings.threat_radius_clear_multiplier

        for id in self.threat_vehicles.keys():
            if self.threat_vehicles[id].distance is not None:
                if self.threat_vehicles[id].distance <= self.ADSB_settings.threat_radius and not self.threat_vehicles[id].is_evading_threat:
                    # if the threat is in the threat radius and not currently
                    # known to the module...
                    # set flag to action threat
                    self.threat_vehicles[id].is_evading_threat = True

                if self.threat_vehicles[id].distance > threat_radius_clear and self.threat_vehicles[id].is_evading_threat:
                    # if the threat is known to the module and outside the
                    # threat clear radius...
                    # clear flag to action threat
                    self.threat_vehicles[id].is_evading_threat = False

        self.active_threat_ids = [id for id in self.threat_vehicles.keys(
        ) if self.threat_vehicles[id].is_evading_threat]

    def update_threat_distances(self, latlonalt):
        '''update the distance between threats and vehicle'''
        for id in self.threat_vehicles.keys():
            threat_latlonalt = (self.threat_vehicles[id].state['lat'] * 1e-7,
                                self.threat_vehicles[id].state['lon'] * 1e-7,
                                self.threat_vehicles[id].state['altitude'])

            self.threat_vehicles[id].h_distance = self.get_h_distance(latlonalt, threat_latlonalt)
            self.threat_vehicles[id].v_distance = self.get_v_distance(latlonalt, threat_latlonalt)

            # calculate and set the total distance between threat and vehicle
            self.threat_vehicles[id].distance = sqrt(
                self.threat_vehicles[id].h_distance**2 + (self.threat_vehicles[id].v_distance)**2)

    def get_h_distance(self, latlonalt1, latlonalt2):
        '''get the horizontal distance between threat and vehicle'''
        (lat1, lon1, alt1) = latlonalt1
        (lat2, lon2, alt2) = latlonalt2

        lat1 = radians(lat1)
        lon1 = radians(lon1)
        lat2 = radians(lat2)
        lon2 = radians(lon2)

        dLat = lat2 - lat1
        dLon = lon2 - lon1

        # math as per mavextra.distance_two()
        a = sin(0.5 * dLat)**2 + sin(0.5 * dLon)**2 * cos(lat1) * cos(lat2)
        c = 2.0 * atan2(sqrt(a), sqrt(1.0 - a))
        return 6371 * 1000 * c

    def get_v_distance(self, latlonalt1, latlonalt2):
        '''get the horizontal distance between threat and vehicle'''
        (lat1, lon1, alt1) = latlonalt1
        (lat2, lon2, alt2) = latlonalt2
        return alt2 - alt1

    def check_threat_timeout(self):
        '''check and handle threat time out'''
        for id in self.threat_vehicles.keys():
            if self.threat_vehicles[id].update_time == 0:
                self.threat_vehicles[id].update_time = self.get_time()
            dt = self.get_time() - self.threat_vehicles[id].update_time
            if dt > self.ADSB_settings.timeout:
                # if the threat has timed out...
                del self.threat_vehicles[id]  # remove the threat from the dict
                for mp in self.module_matching('map*'):
                    # remove the threat from the map
                    mp.map.remove_object(id)
                    mp.map.remove_object(id+":circle")
                # we've modified the dict we're iterating over, so
                # we'll get any more timed-out threats next time we're
                # called:
                return

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == "ADSB_VEHICLE":
            id = 'ADSB-' + str(m.ICAO_address)
            if id not in self.threat_vehicles.keys():  # check to see if the vehicle is in the dict
                # if not then add it
                self.threat_vehicles[id] = ADSBVehicle(id=id, state=m.to_dict())
                for mp in self.module_matching('map*'):
                    from MAVProxy.modules.lib import mp_menu
                    from MAVProxy.modules.mavproxy_map import mp_slipmap
                    self.threat_vehicles[id].menu_item = mp_menu.MPMenuItem(name=id, returnkey=None)

                    threat_radius = get_threat_radius(m)
                    selected_icon = get_threat_icon(m, self.threat_vehicles[id].icon)
                    
                    if selected_icon is not None:
                        # draw the vehicle on the map
                        popup = mp_menu.MPMenuSubMenu('ADSB', items=[self.threat_vehicles[id].menu_item])
                        icon = mp.map.icon(selected_icon)
                        mp.map.add_object(mp_slipmap.SlipIcon(id, (m.lat * 1e-7, m.lon * 1e-7),
                                                    icon, layer=3, rotation=m.heading*0.01, follow=False,
                                                    trail=mp_slipmap.SlipTrail(colour=(0, 255, 255)),
                                                    popup_menu=popup))
                    if threat_radius > 0:
                        mp.map.add_object(mp_slipmap.SlipCircle(id+":circle", 3,
                                                    (m.lat * 1e-7, m.lon * 1e-7),
                                                    threat_radius, (0, 255, 255), linewidth=1))
            else:  # the vehicle is in the dict
                # update the dict entry
                self.threat_vehicles[id].update(m.to_dict(), self.get_time())
                for mp in self.module_matching('map*'):
                    # update the map
                    ground_alt = mp.ElevationMap.GetElevation(m.lat*1e-7, m.lon*1e-7)
                    alt_amsl = m.altitude * 0.001
                    if alt_amsl > 0:
                        alt = int(alt_amsl - ground_alt)
                        label = str(alt) + "m"
                    else:
                        label = None
                    mp.map.set_position(id, (m.lat * 1e-7, m.lon * 1e-7), rotation=m.heading*0.01, label=label, colour=(0,250,250))
                    mp.map.set_position(id+":circle", (m.lat * 1e-7, m.lon * 1e-7))

    def idle_task(self):
        '''called on idle'''
        if self.threat_timeout_timer.trigger():
            self.check_threat_timeout()

        if self.threat_detection_timer.trigger():
            self.perform_threat_detection()
            # TODO: possibly evade detected threats with ids in
            # self.active_threat_ids


def init(mpstate):
    '''initialise module'''
    return ADSBModule(mpstate)



#!/usr/bin/env python
'''
A PPP over MAVLink module
Andrew Tridgell
May 2012
'''

import time, os, fcntl, pty

from MAVProxy.modules.lib import mp_module

class PPPModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(PPPModule, self).__init__(mpstate, "ppp", "PPP link")
        self.command = "noauth nodefaultroute nodetach nodeflate nobsdcomp mtu 128".split()
        self.packet_count = 0
        self.byte_count = 0
        self.ppp_fd = -1
        self.pid = -1
        self.add_command('ppp', self.cmd_ppp, "ppp link control")


    def ppp_read(self, ppp_fd):
        '''called from main select loop in mavproxy when the pppd child
        sends us some data'''
        buf = os.read(ppp_fd, 100)
        if len(buf) == 0:
            # EOF on the child fd
            self.stop_ppp_link()
            return
        print("ppp packet len=%u" % len(buf))
        master = self.master
        master.mav.ppp_send(len(buf), buf)

    def start_ppp_link(self):
        '''startup the link'''
        cmd = ['pppd']
        cmd.extend(self.command)
        (self.pid, self.ppp_fd) = pty.fork()
        if self.pid == 0:
            os.execvp("pppd", cmd)
            raise RuntimeError("pppd exited")
        if self.ppp_fd == -1:
            print("Failed to create link fd")
            return

        # ensure fd is non-blocking
        fcntl.fcntl(self.ppp_fd, fcntl.F_SETFL, fcntl.fcntl(self.ppp_fd, fcntl.F_GETFL) | os.O_NONBLOCK)
        self.byte_count = 0
        self.packet_count = 0

        # ask mavproxy to add us to the select loop
        self.mpself.select_extra[self.ppp_fd] = (self.ppp_read, self.ppp_fd)


    def stop_ppp_link(self):
        '''stop the link'''
        if self.ppp_fd == -1:
            return
        try:
            self.mpself.select_extra.pop(self.ppp_fd)
            os.close(self.ppp_fd)
            os.waitpid(self.pid, 0)
        except Exception:
            pass
        self.pid = -1
        self.ppp_fd = -1
        print("stopped ppp link")


    def cmd_ppp(self, args):
        '''set ppp parameters and start link'''
        usage = "ppp <command|start|stop>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "command":
            if len(args) == 1:
                print("ppp.command=%s" % " ".join(self.command))
            else:
                self.command = args[1:]
        elif args[0] == "start":
            self.start_ppp_link()
        elif args[0] == "stop":
            self.stop_ppp_link()
        elif args[0] == "status":
            self.console.writeln("%u packets %u bytes" % (self.packet_count, self.byte_count))

    def unload(self):
        '''unload module'''
        self.stop_ppp_link()

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == 'PPP' and self.ppp_fd != -1:
            print("got ppp mavlink pkt len=%u" % m.length)
            os.write(self.ppp_fd, m.data[:m.length])

def init(mpstate):
    '''initialise module'''
    return PPPModule(mpstate)



#!/usr/bin/env python

from zipfile import ZipFile
from xml.dom.minidom import parseString

def readkmz(filename):
    '''reads in a kmz file and returns xml nodes'''
    #Strip quotation marks if neccessary
    filename.strip('"')
    #Open the zip file (as applicable)    
    if filename[-4:] == '.kml':
        fo = open(filename, "r")
        fstring = fo.read()
        fo.close()
    elif filename[-4:] == '.kmz':
        zip=ZipFile(filename)
        for z in zip.filelist:
            if z.filename[-4:] == '.kml':
                fstring=zip.read(z)
                break
            else:
                raise Exception("Could not find kml file in %s" % filename)
        else:
            raise Exception("Is not a valid kml or kmz file in %s" % filename)
                
        #send into the xml parser
        kmlstring = parseString(fstring)

        #get all the placenames
        nodes=kmlstring.getElementsByTagName('Placemark')
        
        return nodes
            
def readObject(innode):
    '''reads in a node and returns as a tuple: (type, name, points[])'''
    #get name
    names=innode.getElementsByTagName('name')[0].childNodes[0].data.strip()
        
    #get type
    pointType = 'Unknown'
    if len(innode.getElementsByTagName('LineString')) == 0 and len(innode.getElementsByTagName('Point')) == 0:
        pointType = 'Polygon'
    elif len(innode.getElementsByTagName('Polygon')) == 0 and len(innode.getElementsByTagName('Point')) == 0:
        pointType = 'Polygon'
    elif len(innode.getElementsByTagName('LineString')) == 0 and len(innode.getElementsByTagName('Polygon')) == 0:
        pointType = 'Point'
            
    #get coords
    coords = innode.getElementsByTagName('coordinates')[0].childNodes[0].data.strip()
    coordsSplit = coords.split()
    ret_s = []
    for j in coordsSplit:
        jcoord = j.split(',')
        if len(jcoord) == 3 and jcoord[0] != '' and jcoord[1] != '':
            #print("Got lon " + jcoord[0] + " and lat " + jcoord[1])
            ret_s.append((float(jcoord[1]), float(jcoord[0])))
            
    #return tuple
    return (str(pointType), str(names), ret_s)



class Text():
    '''text to write to console'''
    def __init__(self, text, fg='black', bg='white'):
        self.text = text
        self.fg = fg
        self.bg = bg

class Value():
    '''a value for the status bar'''
    def __init__(self, name, text, row=0, fg='black', bg='white'):
        self.name = name
        self.text = text
        self.row = row
        self.fg = fg
        self.bg = bg


"""
use nokov data to provide motion capture data
it works with nokov software
"""

import time

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
#from MAVProxy.modules.mavproxy_nokov.nokov import nokovsdk

Descriptor_MarkerSet = 0
Descriptor_RigidBody = 1
Descriptor_Skeleton = 2
Descriptor_ForcePlate = 3

nokov_module = None


def py_data_func(pFrameOfMocapData, userdata):
    if pFrameOfMocapData == None:
        print("Not get the data frame.\n")
        return
    frameData = pFrameOfMocapData.contents
    names_rigid = nokov_module.names_rigid
    for i in range(frameData.nRigidBodies):
        if i >= len(names_rigid):
            break
        rigid = frameData.RigidBodies[i]
        name = names_rigid[i]
        if nokov_module.nokov_settings.tracker_name == name:
            print('%f,%f,%f' % (rigid.x, rigid.y, rigid.z))
            now = time.time()
            time_us = int(now * 1.0e6)
            x = rigid.x / 1000
            y = rigid.y / 1000
            z = rigid.z / 1000
            qx = rigid.qx
            qy = rigid.qy
            qz = rigid.qz
            qw = rigid.qw
            if nokov_module.nokov_settings.axis == 'z':
                nokov_module.master.mav.att_pos_mocap_send(time_us, (qw, qy, qx, -qz), y, x, -z)
            elif nokov_module.nokov_settings.axis == 'y':
                nokov_module.master.mav.att_pos_mocap_send(time_us, (qw, qx, qz, -qy), x, z, -y)
            return


class NokovModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(NokovModule, self).__init__(mpstate, "nokov", "nokov")
        global nokov_module
        nokov_module = self
        self.client = None
        self.names_rigid = []
        self.nokov_settings = mp_settings.MPSettings(
            [('host', str, '127.0.0.1'),
             ('axis', str, 'z'),
             ('tracker_name', str, None)]
        )
        self.add_command('nokov', self.cmd_nokov, "nokov control", ['<start>', '<stop>', 'set (NOKOVSETTING)'])

    def cmd_stop(self):
        del self.client
        self.client = None
        self.names_rigid = []

    def cmd_start(self):
        if self.client != None:
            print('The connection already exists, please disconnect first(command stop).')
            return
        print('serverIp is %s' % self.nokov_settings.host)
        client = nokovsdk.PySDKClient()
        ver = client.PySeekerVersion()
        print('SeekerSDK ver. %d.%d.%d.%d' % (ver[0], ver[1], ver[2], ver[3]))
        client.PySetDataCallback(py_data_func, None)
        ret = client.Initialize(bytes(self.nokov_settings.host, encoding="utf8"))
        if ret == 0:
            print("Connect to the Seeker Succeed")
            dsc = nokovsdk.DataDescriptions()
            handle = nokovsdk.c_void_p()
            ret = client.PyGetDataDescriptionsEx(dsc, handle)
            if ret == 0:
                print("GetDataDescriptions Succeed")
                self.parseDescriptions(dsc)
                self.client = client
                client.PyFreeDataDescriptionsEx(handle)
            else:
                print("GetDataDescriptions Failed: [%d]" % ret)
                self.cmd_stop()
        else:
            print("Connect Failed: [%d]" % ret)
            self.cmd_stop()

    def parseDescriptions(self, dsc):
        for i in range(dsc.nDataDescriptions):
            dc = dsc.arrDataDescriptions[i]
            if dc.type == Descriptor_RigidBody:
                name = str(dc.Data.RigidBodyDescription.contents.szName, encoding="utf8")
                self.names_rigid.append(name)
                print('Descriptions rigid:', name)

    def usage(self):
        '''show help on command line options'''
        return "Usage: nokov <start|stop|set>"

    def cmd_nokov(self, args):
        '''control behaviour of the module'''
        if len(args) == 0:
            print(self.usage())
        elif args[0] == "start":
            self.cmd_start()
        elif args[0] == "stop":
            self.cmd_stop()
        elif args[0] == "set":
            self.nokov_settings.command(args[1:])
        else:
            print(self.usage())

    def idle_task(self):
        '''called rapidly by mavproxy'''
        pass


def init(mpstate):
    '''initialise module'''
    return NokovModule(mpstate)



'''
GraphDefinition class
'''

class GraphDefinition(object):
    '''a pre-defined graph'''
    def __init__(self, name, expression, description, expressions, filename):
        self.name = name
        self.expression = expression
        self.description = description
        self.expressions = expressions
        self.filename = filename



#!/usr/bin/env python
'''auxopt command handling'''

import time, os
from MAVProxy.modules.lib import mp_module


aux_options = {
    "Nothing":"0",
    "Flip":"2",
    "SimpleMode":"3",
    "RTL":"4",
    "SaveTrim":"5",
    "SaveWP":"7",
    "MultiMode":"8",
    "CameraTrigger":"9",
    "Sonar":"10",
    "Fence":"11",
    "ResetYaw":"12",
    "SuperSimpleMode":"13",
    "AcroTrainer":"14",
    "Auto":"16",
    "AutoTune":"17",
    "Land":"18"
}

class AuxoptModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(AuxoptModule, self).__init__(mpstate, "auxopt", "auxopt command handling")
        self.add_command('auxopt', self.cmd_auxopt,   'select option for aux switches on CH7 and CH8 (ArduCopter only)',
                         ['set <7|8> <Nothing|Flip|SimpleMode|RTL|SaveTrim|SaveWP|MultiMode|CameraTrigger|Sonar|Fence|ResetYaw|SuperSimpleMode|AcroTrainer|Acro|Auto|AutoTune|Land>',
                          'reset <7|8|all>',
                          '<show|list>'])

    def aux_show(self, channel):
        param = "CH%s_OPT" % channel
        opt_num = str(int(self.get_mav_param(param)))
        option = None
        for k in aux_options.keys():
            if opt_num == aux_options[k]:
                option = k
                break
        else:
            print("AUX Channel is currently set to unknown value " + opt_num)
            return
        print("AUX Channel is currently set to " + option)

    def aux_option_validate(self, option):
        for k in aux_options:
            if option.upper() == k.upper():
                return k
        return None

    def cmd_auxopt(self, args):
        '''handle AUX switches (CH7, CH8) settings'''
        if self.mpstate.vehicle_type != 'copter':
            print("This command is only available for copter")
            return
        if len(args) == 0 or args[0] not in ('set', 'show', 'reset', 'list'):
            print("Usage: auxopt set|show|reset|list")
            return
        if args[0] == 'list':
            print("Options available:")
            for s in sorted(aux_options.keys()):
                print('  ' + s)
        elif args[0] == 'show':
            if len(args) > 2 and args[1] not in ['7', '8', 'all']:
                print("Usage: auxopt show [7|8|all]")
                return
            if len(args) < 2 or args[1] == 'all':
                self.aux_show('7')
                self.aux_show('8')
                return
            self.aux_show(args[1])
        elif args[0] == 'reset':
            if len(args) < 2 or args[1] not in ['7', '8', 'all']:
                print("Usage: auxopt reset 7|8|all")
                return
            if args[1] == 'all':
                self.param_set('CH7_OPT', '0')
                self.param_set('CH8_OPT', '0')
                return
            param = "CH%s_OPT" % args[1]
            self.param_set(param, '0')
        elif args[0] == 'set':
            if len(args) < 3 or args[1] not in ['7', '8']:
                print("Usage: auxopt set 7|8 OPTION")
                return
            option = self.aux_option_validate(args[2])
            if not option:
                print("Invalid option " + args[2])
                return
            param = "CH%s_OPT" % args[1]
            self.param_set(param, aux_options[option])
        else:
            print("Usage: auxopt set|show|list")

def init(mpstate):
    '''initialise module'''
    return AuxoptModule(mpstate)



'''
Graphical editing of mp_settings object
'''
import os, sys
from MAVProxy.modules.lib import multiproc
import threading

class WXSettings(object):
    '''
    a graphical settings dialog for mavproxy
    '''
    def __init__(self, settings):
        self.settings  = settings
        self.parent_pipe,self.child_pipe = multiproc.Pipe()
        self.close_event = multiproc.Event()
        self.close_event.clear()
        self.child = multiproc.Process(target=self.child_task)
        self.child.start()
        t = threading.Thread(target=self.watch_thread)
        t.daemon = True
        t.start()

    def child_task(self):
        '''child process - this holds all the GUI elements'''
        from MAVProxy.modules.lib import mp_util
        from MAVProxy.modules.lib import wx_processguard
        from MAVProxy.modules.lib.wx_loader import wx
        from MAVProxy.modules.lib.wxsettings_ui import SettingsDlg

        mp_util.child_close_fds()
        app = wx.App(False)
        dlg = SettingsDlg(self.settings)
        dlg.parent_pipe = self.parent_pipe
        dlg.ShowModal()
        dlg.Destroy()

    def watch_thread(self):
        '''watch for settings changes from child'''
        from MAVProxy.modules.lib.mp_settings import MPSetting
        while True:
            setting = self.child_pipe.recv()
            if not isinstance(setting, MPSetting):
                break
            try:
                self.settings.set(setting.name, setting.value)
            except Exception:
                print("Unable to set %s to %s" % (setting.name, setting.value))

    def is_alive(self):
        '''check if child is still going'''
        return self.child.is_alive()


if __name__ == "__main__":
    multiproc.freeze_support()

    def test_callback(setting):
        '''callback on apply'''
        print("Changing %s to %s" % (setting.name, setting.value))

    # test the settings
    from MAVProxy.modules.lib import mp_settings, time
    from MAVProxy.modules.lib.mp_settings import MPSetting
    settings = mp_settings.MPSettings(
        [ MPSetting('link', int, 1, tab='TabOne'),
          MPSetting('altreadout', int, 10, range=(-30,1017), increment=1),
          MPSetting('pvalue', float, 0.3, range=(-3.0,1e6), increment=0.1, digits=2),
          MPSetting('enable', bool, True, tab='TabTwo'),
          MPSetting('colour', str, 'Blue', choice=['Red', 'Green', 'Blue']),
          MPSetting('foostr', str, 'blah', label='Foo String') ])
    settings.set_callback(test_callback)
    dlg = WXSettings(settings)
    while dlg.is_alive():
        time.sleep(0.1)
